import json
import random
from datetime import datetime
import requests
import boto3
from botocore.config import Config
from .app_mxnet_inference import get_dma
from .utils_ncf import get_ncf_ranked_offers


def get_assigned_offers(uid, rest_tz):
    my_config = Config(region_name='us-east-1', connect_timeout=2, read_timeout=2, retries={'max_attempts': 2})
    dynamodb = boto3.resource('dynamodb', config=my_config)
    table = dynamodb.Table("digitalGuestProfile")
    try:
        response = table.get_item(Key={'userId': uid})
        if response.get("ResponseMetadata", {}).get("HTTPStatusCode", "") == 200:
            if "Item" in response and "assignedOffers" in response["Item"]:
                get_assigned_offers_ind = "success"
                assigned_offers = response["Item"].get("assignedOffers", {})
                user_location = response['Item'].get('user_attributes', {}).get('location', {})
                # get loyalty_id
                loyalty_id = response["Item"].get("loyaltyId")
                if "purchases" in response["Item"]:
                    purchase_history = response['Item'].get('purchases', [])
                else:
                    purchase_history = []
            elif "Item" not in response:
                user_location = {}
                purchase_history = []
                get_assigned_offers_ind = "userNotExisted"
                assigned_offers = {}
                loyalty_id = None
            elif "assignedOffers" not in response.get("Item", []):
                user_location = {}
                purchase_history = []
                get_assigned_offers_ind = "userNoAssignedOffers"
                assigned_offers = {}
                loyalty_id = None
            else:
                user_location = {}
                purchase_history = []
                get_assigned_offers_ind = "implementError"
                assigned_offers = {}
                loyalty_id = None
        else:
            user_location = {}
            purchase_history = []
            get_assigned_offers_ind = response.get("ResponseMetadata", {}).get("HTTPStatusCode", "")
            assigned_offers = {}
            loyalty_id = None
    except Exception as e:
        user_location = {}
        purchase_history = []
        get_assigned_offers_ind = str(e)
        assigned_offers = {}
        loyalty_id = None
    store_history = [str(purchase.get("bkid", "0")) for purchase in purchase_history]
    user_tier, is_phoenix = get_dma(store_ids=store_history, user_location=user_location, dim_rest=rest_tz)
    return purchase_history, assigned_offers, get_assigned_offers_ind, user_tier, loyalty_id


def get_daily_offer(offer_pools, offer_attributes, uid, user_tier, assigned_offers, cluster, lapsed_category, daily_personalized_k,
                    daily_lapsed_k, daily_offer_id, daily_lapsed_id, offer_group_lookup, daily_offer_id_backup):
    dt_now = datetime.utcnow()
    daily_expires_on = assigned_offers.get("dailyOffersExpireOn", "2000-01-01T00:00:00")
    dt_old_daily_expires = datetime.fromisoformat(daily_expires_on.split(".")[0])
    is_newly_assigned_daily = (dt_old_daily_expires - dt_now).total_seconds() > 3600 * 12  # 12 hours
    assigned_offer_groups = [
        offer_group_lookup.get(offer_id, "0")
        for offer_id in assigned_offers.get("weeklyOffers", [])
    ]
    if not is_newly_assigned_daily:
        price_tier = "1_Personalized" if user_tier == "1" else "2_Personalized"
        # rank offers - randomly
        daily_offer_ranked_random = random.sample(daily_offer_id, len(daily_offer_id))
        daily_lapsed_ranked_random = random.sample(daily_lapsed_id, len(daily_lapsed_id))
        daily_offer_sup_random = random.sample(daily_offer_id_backup[user_tier], len(daily_offer_id_backup[user_tier]))

        # rank offers - NCF
        daily_offer_ranked_ncf, get_assigned_offers_ind_ncf, user_group_ncf = get_ncf_ranked_offers(
            rbi_cognito_id=uid,
            table_name="bk-us-ncf-pred"
        )
        if get_assigned_offers_ind_ncf == "success" and user_group_ncf == "test":
            daily_offer_ranked = [
                oid
                for oid in daily_offer_ranked_ncf
                if oid in offer_pools["weekly"]
                and offer_attributes.get(oid, {}).get("Offer Type", "") == price_tier
            ]
            daily_lapsed_ranked = [
                oid
                for oid in daily_offer_ranked_ncf
                if oid in offer_pools["weekly"]
                if offer_attributes.get(oid, {}).get("Offer Type", "") in ["Lapsed_60", "Lapsed_90"]
            ]
        else:
            daily_offer_ranked = daily_offer_ranked_random
            daily_lapsed_ranked = daily_lapsed_ranked_random
        daily_offer_ranked += daily_offer_sup_random

        # pick ranked offers
        selected_personalized_offers = []
        for offer_id in daily_offer_ranked:
            offer_group = offer_group_lookup.get(offer_id, "0")
            if offer_group == "0" or offer_group not in assigned_offer_groups:
                selected_personalized_offers.append(offer_id)
                assigned_offer_groups.append(offer_group)
                if len(selected_personalized_offers) >= daily_personalized_k:
                    break
        selected_lapsed_offers = []
        if lapsed_category in ["60", "90"]:
            for offer_id in daily_lapsed_ranked:
                offer_group = offer_group_lookup.get(offer_id, "0")
                if offer_group == "0" or offer_group not in assigned_offer_groups:
                    selected_lapsed_offers.append(offer_id)
                    assigned_offer_groups.append(offer_group)
                    if len(selected_lapsed_offers) >= daily_lapsed_k:
                        break
        selected_offers = selected_personalized_offers + selected_lapsed_offers
    else:
        selected_offers = assigned_offers.get("dailyOffers", ["n/a"])
    return selected_offers, assigned_offer_groups


def update_profile_offers(uid, assigned_offers):
    my_config = Config(region_name='us-east-1', connect_timeout=2, read_timeout=2, retries={'max_attempts': 2})
    dynamodb = boto3.resource('dynamodb', config=my_config)
    table = dynamodb.Table("digitalGuestProfile")

    update_expression = "set #attrName = :attrValue"
    expr_attr_names = {"#attrName": "assignedOffers"}
    expr_attr_values = {":attrValue": assigned_offers}
    # if loyalty_id:
    #     update_expression += ", #attrNameA = :attrValue1"
    #     expr_attr_names.update({"#attrNameA": "loyaltyId"})
    #     expr_attr_values.update({":attrValue1": loyalty_id})
    # if user_attributes:
    #     update_expression += ", #attrNameB = :attrValue2"
    #     expr_attr_names.update({"#attrNameB": "user_attributes"})
    #     expr_attr_values.update({":attrValue2": user_attributes})
    try:
        update_response = table.update_item(
            Key={
                'userId': uid
            },
            UpdateExpression=update_expression,
            ExpressionAttributeNames=expr_attr_names,
            ExpressionAttributeValues=expr_attr_values,
            ReturnValues="UPDATED_NEW",
        )
        status_code = update_response['ResponseMetadata']['HTTPStatusCode']
    except Exception as e:
        status_code = str(e)
    return status_code


def update_amplitude_event(uid, attributes_d):
    amplitude_data = {
        "events":
            [
                {
                    "data": {
                        "custom_event_type": "user_preference",
                        "event_name": "Daily Offer Personalization",
                        "custom_attributes": attributes_d
                    },
                    "event_type": "custom_event"
                }
            ],
        "environment": "production",
        "user_identities": {"customer_id": uid}
    }
    amplitude_json_data = json.dumps(amplitude_data)
    amplitude_response = requests.post('https://s2s.mparticle.com/v2/events',
                                       auth=requests.auth.HTTPBasicAuth(
                                           'us1-9e24f161c6023f4e8e2d9a52ad8c6b37',
                                           'gplReWa5XTlIQ5_zO_78p0X4habcNjfaj-N6xDipT3ayp9jGJevvyI52pD2_n5gK'),
                                       data=amplitude_json_data)
    try:
        amplitude_response_code = str(amplitude_response.status_code)
    except Exception as e:
        amplitude_response_code = 'exceptionError'
    return amplitude_response_code


def draw_national_group(national_group_weights):
    group_names = national_group_weights.get("groups")
    weights = national_group_weights.get("weights")
    assert len(group_names) == len(weights), "incomparable length groups and weights"

    group = random.choices(population=group_names, weights=weights)[0]
    return group



import math
import numpy as np
import mxnet as mx
from gluonnlp.data import UnigramCandidateSampler
from gluonnlp.model import ISDense
from gluonnlp.model.transformer import TransformerEncoder, TransformerEncoderCell
from mxnet import ndarray, gluon


class FrozenEmbedding(gluon.nn.Embedding):
    def __init__(self, frozen_weight, **kwargs):
        input_dim, output_dim = frozen_weight.shape
        super(FrozenEmbedding, self).__init__(input_dim=input_dim, output_dim=output_dim, **kwargs)
        self.frozen_weight = frozen_weight

    def initialize_frozen_weight(self, ctx=None, frozen_weight=None, freeze=True):
        self.initialize(force_reinit=True)
        if frozen_weight is not None:
            self.frozen_weight = frozen_weight
        self.weight.set_data(self.frozen_weight)
        if ctx is not None:
            self.weight.reset_ctx(ctx=ctx)
        if freeze:
            for param in self.collect_params().values():
                param.grad_req = "null"


class MeanMaxPooling(gluon.nn.HybridBlock):
    def __init__(self, axis=1, dropout=0.0, prefix=None, params=None, **kwargs):
        super().__init__(prefix=prefix, params=params)
        self.axis = axis
        self.dropout = gluon.nn.Dropout(rate=dropout)

    def hybrid_forward(self, F, inputs, valid_length=None):
        mean_out = F.mean(data=inputs, axis=self.axis) if valid_length is None \
            else F.sum(data=inputs, axis=self.axis) / valid_length
        max_out = F.max(data=inputs, axis=self.axis)
        outputs = F.concat(mean_out, max_out, dim=1)
        if self.dropout:
            outputs = self.dropout(outputs)
        #         outputs = F.LayerNorm(outputs)
        return outputs


class SequenceVAE(gluon.HybridBlock):
    def __init__(self, n_hidden=400, n_latent=100, n_layers=1, n_output=768+2048, batch_size=100,
                 act_type='relu', ctx=mx.cpu(), **kwargs):
        self.soft_zero = 1e-10
        self.n_latent = n_latent
        self.batch_size = batch_size
        self.model_ctx = ctx
        self.output = None
        self.mu = None
        # note to self: requring batch_size in model definition is sad, not sure how to deal with this otherwise though
        super(SequenceVAE, self).__init__(**kwargs)
        # self.use_aux_logits = use_aux_logits
        with self.name_scope():
            self.encoder = gluon.nn.HybridSequential(prefix='encoder')
            for i in range(n_layers):
                self.encoder.add(gluon.nn.Dense(n_hidden, activation=act_type))
            self.encoder.add(gluon.nn.Dense(n_latent * 2, activation=None))

            self.decoder = gluon.nn.HybridSequential(prefix='decoder')
            for i in range(n_layers):
                self.decoder.add(gluon.nn.Dense(n_hidden, activation=act_type))
            self.decoder.add(gluon.nn.Dense(n_output, activation='sigmoid'))

    def hybrid_forward(self, F, x):
        h = self.encoder(x)
        # print(h)
        mu_lv = F.split(h, axis=1, num_outputs=2)
        mu = mu_lv[0]
        lv = mu_lv[1]
        self.mu = mu
        # eps = F.random_normal(loc=0, scale=1, shape=mu.shape, ctx=model_ctx)
        # this would work fine only for nd (i.e. non-hybridized block)
        eps = F.random_normal(loc=0, scale=1, shape=(self.batch_size, self.n_latent),
                              ctx=self.model_ctx)
        z = mu + F.exp(0.5 * lv) * eps
        y = self.decoder(z)
        self.output = y

        KL = 0.5 * F.sum(1 + lv - mu * mu - F.exp(lv), axis=1)
        logloss = F.sum(x * F.log(y + self.soft_zero) + (1 - x) * F.log(1 - y + self.soft_zero), axis=1)
        loss = -logloss - KL

        return loss, z, y

    def initialize_frozen(self, param_path, batch_size, ctx):
        self.initialize(force_reinit=True)
        self.load_parameters(param_path, ctx=ctx)
        for param in self.collect_params().values():
            param.grad_req = "null"
        self.set_model_ctx(ctx=ctx)
        self.set_batch_size(batch_size=batch_size)

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size

    def set_model_ctx(self, ctx):
        self.model_ctx = ctx


class SequenceTransformer(gluon.nn.HybridBlock):
    def __init__(self, num_items, item_embed, item_hidden_size, item_max_length, item_num_heads,
                 item_num_layers, item_transformer_dropout, item_pooling_dropout, cross_size, item_embedding_weight,
                 prefix=None, params=None, **kwargs):
        super().__init__(prefix=prefix, params=params)
        with self.name_scope():
            if item_embedding_weight is None:
                self.item_embedding = gluon.nn.Embedding(input_dim=num_items, output_dim=item_embed)
            else:
                self.item_embedding = FrozenEmbedding(frozen_weight=item_embedding_weight)
            self.item_encoder = TransformerEncoder(units=item_embed,
                                                   hidden_size=item_hidden_size,
                                                   num_heads=item_num_heads,
                                                   num_layers=item_num_layers,
                                                   max_length=item_max_length,
                                                   dropout=item_transformer_dropout)
            self.item_pooling_dp = MeanMaxPooling(dropout=item_pooling_dropout)
            self.dense = gluon.nn.Dense(cross_size)

    def hybrid_forward(self, F, input_item, item_valid_length=None, input_history=None):
        item_embed_out = self.item_embedding(input_item)
        if input_history is not None:
            history_embed_out = self.item_embedding(input_history)
            history_embed_avg = history_embed_out.mean(axis=1, keepdims=True)
            item_embed_out = F.concat(history_embed_avg, item_embed_out, dim=1)
            item_valid_length = F.broadcast_add(item_valid_length, F.ones_like(item_valid_length))
            # item_valid_length = mx.symbol.broadcast_add(item_valid_length, 1)
        item_encoding, item_att = self.item_encoder(inputs=item_embed_out, valid_length=item_valid_length)
        item_out = self.item_pooling_dp(item_encoding)
        item_out = self.dense(item_out)

        return item_out


class ContextTransformer(gluon.nn.HybridBlock):
    def __init__(self, context_dims, context_embed, context_hidden_size,
                 context_num_heads, context_transformer_dropout, context_pooling_dropout,
                 cross_size, prefix=None, params=None, **kwargs):
        super().__init__(prefix=prefix, params=params)
        self.context_dims = context_dims
        self.context_embed = context_embed
        self.cross_size = cross_size
        with self.name_scope():
            self.context_pooling_dp = MeanMaxPooling(dropout=context_pooling_dropout)
            self.context_encoder = TransformerEncoderCell(units=context_embed,
                                                          hidden_size=context_hidden_size,
                                                          num_heads=context_num_heads,
                                                          dropout=context_transformer_dropout
                                                          )
            self.dense = gluon.nn.Dense(self.cross_size)
            self.context_embeddings = gluon.nn.HybridSequential()
            for i, context_dim in enumerate(self.context_dims):
                self.context_embeddings.add(gluon.nn.Embedding(self.context_dims[i], self.context_embed))

    def hybrid_forward(self, F, input_context_list):
        context_embed = [self.context_embeddings[i](input_context) for i, input_context in enumerate(input_context_list)]
        context_input = []
        for i in context_embed:
            context_input.append(F.expand_dims(i, axis=1))
        context_embedding = F.concat(*context_input, dim=1)
        context_encoding, context_att = self.context_encoder(context_embedding)
        context_out = self.context_pooling_dp(context_encoding)
        context_out = self.dense(context_out)

        return context_out


class TxT(gluon.nn.HybridBlock):
    def __init__(self,
                 num_items,
                 context_dims,
                 item_embed=200,
                 context_embed=100,
                 item_hidden_size=512,
                 item_max_length=5,
                 item_num_heads=4,
                 item_num_layers=2,
                 item_transformer_dropout=0.0,
                 item_pooling_dropout=0.0,
                 context_hidden_size=512,
                 context_num_heads=2,
                 context_transformer_dropout=0.0,
                 context_pooling_dropout=0.0,
                 act_type="gelu",
                 cross_size=200,
                 item_embedding_weight=None,
                 prefix=None,
                 params=None,
                 **kwargs):
        super(TxT, self).__init__(**kwargs)
        self.act_type = act_type
        with self.name_scope():
            self.sequence_transformer = SequenceTransformer(
                num_items=num_items,
                item_embed=item_embed,
                item_hidden_size=item_hidden_size,
                item_max_length=item_max_length,
                item_num_heads=item_num_heads,
                item_num_layers=item_num_layers,
                item_transformer_dropout=item_transformer_dropout,
                item_pooling_dropout=item_pooling_dropout,
                cross_size=cross_size,
                item_embedding_weight=item_embedding_weight,
                prefix=prefix,
                params=params,
            )
            self.context_transformer = ContextTransformer(
                context_dims=context_dims,
                context_embed=context_embed,
                context_hidden_size=context_hidden_size,
                context_num_heads=context_num_heads,
                context_transformer_dropout=context_transformer_dropout,
                context_pooling_dropout=context_pooling_dropout,
                cross_size=cross_size,
                prefix=prefix,
                params=params,
            )
            self.dense1 = gluon.nn.Dense(units=num_items//2)
            if act_type == "relu":
                self.act = gluon.nn.Activation(activation="relu")
            elif act_type == "gelu":
                self.act = gluon.nn.GELU()
            elif act_type == "leakyRelu":
                self.act = gluon.nn.LeakyReLU(alpha=0.2)
            else:
                raise NotImplementedError
            self.dense2 = gluon.nn.Dense(units=num_items, activation=None)

    def hybrid_forward(self, F, input_item, item_valid_length, input_context_list, input_history=None):
        item_outs = self.sequence_transformer(input_item, item_valid_length, input_history)
        context_outs = self.context_transformer(input_context_list)

        outs = F.broadcast_mul(item_outs, context_outs)
        outs = self.dense1(outs)
        outs = self.act(outs)
        outs = self.dense2(outs)

        return outs



import json
import random
import pandas as pd


class NationalOffer:
    def __init__(self, offer):
        group = offer.get("group", "n/a")
        sanity_id = offer.get("sanity_id", "n/a")
        name = offer.get("name", "n/a")
        tier = offer.get("tier", "n/a")
        lapsed = offer.get("lapsed", "n/a")
        cluster = offer.get("cluster", "n/a")
        test_group_negate = offer.get("test_group_negate", "n/a").split("|")
        prob = offer.get("prob", 0)
        all_tier = ["1_Personalized", "2_Personalized"]
        all_lapsed = ["0", "30", "60", "90"]
        all_cluster = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
        assert tier in all_tier + ["all"], "Incorrect tier value"
        assert lapsed in all_lapsed + ["all"], "Incorrect lapsed value"
        assert cluster in all_cluster + ["all"], "Incorrect cluster value"
        assert 0 < prob <= 1, "Incorrect prob value"
        self.group = group
        self.sanity_id = sanity_id
        self.name = name
        self.tier = tier
        self.lapsed = lapsed
        self.cluster = cluster
        self.test_group_negate = test_group_negate
        self.prob = prob

    def __repr__(self):
        return json.dumps({
            "group": self.group,
            "sanity_id": self.sanity_id,
            "name": self.name,
            "tier": self.tier,
            "lapsed": self.lapsed,
            "cluster": self.cluster,
            "test_group_negate": self.test_group_negate,
            "prob": self.prob,
        })

    def satisfy_rule(self, tier=None, lapsed=None, cluster=None, test_group=None):
        return True if all([
            self.is_tier(tier),
            self.has_lapsed(lapsed),
            self.has_cluster(cluster),
            self.is_test_group(test_group),
        ]) else False

    def is_tier(self, tier):
        return True if tier is None or self.tier == "all" or tier == self.tier else False

    def has_lapsed(self, lapsed):
        return True if lapsed is None or self.lapsed == "all" or lapsed == self.lapsed else False

    def has_cluster(self, cluster):
        return True if cluster is None or self.cluster == "all" or cluster == self.cluster else False

    def is_test_group(self, test_group):
        return True if test_group is None \
            or "none" in self.test_group_negate \
            or test_group not in self.test_group_negate \
            else False


class NationalOfferGroup:
    def __init__(self, group):
        self.group = self.load_dict(group)
        self.num = len(self.group)

    def pick(self, tier=None, lapsed=None, cluster=None, test_group=None):
        offers = [
            offer for offer in self.group
            if offer.satisfy_rule(tier=tier, lapsed=lapsed, cluster=cluster, test_group=test_group)
        ]
        probs = [offer.prob for offer in offers]
        if len(offers) > 0:
            selected_offer = random.choices(population=offers, weights=probs, k=1)[0]
            selected_id = selected_offer.sanity_id
            selected_name = selected_offer.name
        else:
            selected_id = "n/a"
            selected_name = "n/a"
        return selected_id, selected_name

    def list_offers(self, attr="id", tier=None, lapsed=None, cluster=None, test_group=None):
        offers = [
            offer for offer in self.group
            if offer.satisfy_rule(tier=tier, lapsed=lapsed, cluster=cluster, test_group=test_group)
        ]
        if attr == "id" or attr == "sanity_id":
            offer_attr = list(set([offer.sanity_id for offer in offers]))
        elif attr == "name":
            offer_attr = list(set([offer.name for offer in offers]))
        else:
            raise NotImplementedError()
        return offer_attr

    def load_dict(self, group):
        return [
            NationalOffer(offer=offer)
            for offer in group
        ]


class NationalOfferPool:
    def __init__(self, pool, source="csv"):
        if source == "csv":
            self.pool = self.load_csv(pool)
        elif source == "dict":
            self.pool = self.load_dict(pool)
        else:
            raise NotImplementedError()
        self.max_num = len(self.pool)

    def pick(self, tier=None, lapsed=None, cluster=None, test_group=None):
        return [
            group.pick(tier=tier, lapsed=lapsed, cluster=cluster, test_group=test_group)
            for group in self.pool
        ]

    def list_offers(self, attr="id", tier=None, lapsed=None, cluster=None, test_group=None):
        return [
            offer
            for group in self.pool
            for offer in group.list_offers(attr=attr, tier=tier, lapsed=lapsed, cluster=cluster, test_group=test_group)
        ]

    def load_csv(self, pool):
        df = pd.read_csv(pool)
        assert all(df.groupby(by=["group", "tier", "lapsed", "cluster", "test_group_negate"]).sum("prob").values == 1)
        assert not df.isna().any().any()
        all_groups = df["group"].unique()
        dict_pool = [
            df[df["group"] == group].to_dict(orient="records")
            for group in all_groups
        ]
        return self.load_dict(pool=dict_pool)

    def load_dict(self, pool):
        return [
            NationalOfferGroup(group=group)
            for group in pool
        ]

# class PromotionalOffer:
#     def __init__(self, offer):
#
#
# class PromotionalOfferPool:



from botocore.config import Config
import boto3


def get_ncf_ranked_offers(rbi_cognito_id, table_name="bk-us-ncf-pred"):
    my_config = Config(region_name='us-east-1', connect_timeout=2, read_timeout=2, retries={'max_attempts': 2})
    dynamodb = boto3.resource('dynamodb', config=my_config)
    ncf_pred_table = dynamodb.Table(table_name)
    try:
        response = ncf_pred_table.get_item(Key={'rbi_cognito_id': rbi_cognito_id})
        if response.get("ResponseMetadata", {}).get("HTTPStatusCode", "") == 200:
            if "Item" in response and "ranked_offers" in response["Item"] and "group" in response["Item"]:
                user_group = response["Item"]["group"]
                ranked_offers = response["Item"]["ranked_offers"]
                get_assigned_offers_ind = "success"
            elif "Item" not in response:
                ranked_offers = []
                user_group = ""
                get_assigned_offers_ind = "userNotInNCF"
            elif "ranked_offers" not in response.get("Item", []):
                ranked_offers = []
                user_group = ""
                get_assigned_offers_ind = "noRankedOffer"
            elif "user_group" not in response.get("Item", []):
                ranked_offers = []
                user_group = ""
                get_assigned_offers_ind = "noGroupAssigned"
            else:
                ranked_offers = []
                user_group = ""
                get_assigned_offers_ind = "implementError"
        else:
            ranked_offers = []
            user_group = ""
            get_assigned_offers_ind = response.get("ResponseMetadata", {}).get("HTTPStatusCode", "")
    except Exception as e:
        ranked_offers = []
        user_group = ""
        get_assigned_offers_ind = str(e)
    return ranked_offers, get_assigned_offers_ind, user_group



from .callbacks_base import TensorboardCallbackBase


class MTTensorboardCallback(TensorboardCallbackBase):
    def __init__(self, tensorboard_logger):
        super().__init__(tensorboard_logger)

    def update_metrics_scalar(self, trainer, pl_module, phase):
        epoch = pl_module.current_epoch
        if phase == "Train":
            self.tensorboard_logger.experiment.add_scalar(
                f"loss/{phase}", trainer.callback_metrics["train_loss"], epoch
            )
        elif phase == "Validation":
            self.tensorboard_logger.experiment.add_scalar(
                f"loss/{phase}", trainer.callback_metrics["val_loss"], epoch
            )



import os
import pandas as pd
import numpy as np
import mxnet as mx
from mxnet import nd, autograd, gluon
import boto3


# Helper Functions - Models

# Should be changed along with utils_offer_encoder.load_offer_id_from_pool
def load_offer_id_from_pool(offer_pool_path, active_offer_pool=None):
    offer_pool = pd.read_csv(offer_pool_path, dtype="str").dropna(subset=["Sanity ID"])
    offer_pool = offer_pool[~offer_pool["Sanity ID"].isin(["TBR", "TBD", "0", "5"])]
    offer_pool = offer_pool[~offer_pool["Sanity ID"].str.lower().str.contains("requested")]
    print("Offer Pool without TBD and NA: %d" % len(offer_pool))

    # offer pools
    weekly_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &  # column removed
        # (offer_pool["National Offer Overlap?"] == "No") &  # column removed
        (offer_pool["Offer Type"].isin(["1_Personalized", "2_Personalized", "Lapsed_60", "Lapsed_90", "Default"]))
        # (offer_pool["L3"].isin(["Lapsed", "Segmented"])) &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    weekly_default_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &
        # (offer_pool["National Offer Overlap?"] == "No") &
        (offer_pool["Offer Type"] == "Default")
        # (offer_pool["L3"].isin(["Lapsed", "Segmented"])) &
        # (offer_pool["L4"].isin(["Default"]))
        ]["Sanity ID"].tolist()
    daily_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &
        # (offer_pool["National Offer Overlap?"] == "No") &
        (offer_pool["Offer Type"] == "Daily")
        # (offer_pool["L3"] == "Daily Deals") &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    delivery_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "Delivery") &
        (offer_pool["Offer Type"] == "Delivery")
        # (offer_pool["L3"].isin(["Lapsed", "Segmented", "Daily Deals"])) &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    if active_offer_pool:
        weekly_offer_pool = [offer_id for offer_id in weekly_offer_pool if offer_id in active_offer_pool]
        weekly_default_pool = [offer_id for offer_id in weekly_default_pool if offer_id in active_offer_pool]
        daily_offer_pool = [offer_id for offer_id in daily_offer_pool if offer_id in active_offer_pool]
        delivery_offer_pool = [offer_id for offer_id in delivery_offer_pool if offer_id in active_offer_pool]
    inactive_pool = offer_pool[
        offer_pool["Sanity ID"].isin([i for i in offer_pool["Sanity ID"].values
                                      if i not in weekly_offer_pool + daily_offer_pool + delivery_offer_pool])
    ]
    print("active weekly MO&P: %d" % len(weekly_offer_pool))
    print("active weekly MO&P default: %d" % len(weekly_default_pool))
    print("active daily MO&P: %d" % len(daily_offer_pool))
    print("active weekly Delivery: %d" % len(delivery_offer_pool))
    offer_pools = {
        "weekly": weekly_offer_pool,
        "default": weekly_default_pool,
        "daily": daily_offer_pool,
        "delivery": delivery_offer_pool,
        "inactive": inactive_pool,
    }

    # offer attributes
    offer_attributes = offer_pool[
        ["Sanity ID", "Coupon Name", "Offer Type", "Offer Structure", "Product Category", "Group Name",
         "Discount Level", "Personalized Offer Group"]
    ].set_index(keys="Sanity ID").fillna("NA").to_dict(orient="index")

    return offer_pools, offer_attributes


def generate_lookup_dict(df, key_col, value_col):
    """Generate int2int lookup dict from given dataframe."""
    return df[[key_col, value_col]].set_index(key_col).to_dict()[value_col]


def transform_raw_to_history(history_raw):
    """
    Transform user raw_history to processed recent history
    :param history_raw: list of user offer redeemed events
    :return: processed recent history of the user
    """
    # filtered_history = [v for v in raw_history if v.get("orderType") == "restaurant order"]
    sorted_history = sorted(history_raw, key=lambda v: v.get("timeStamp", "0000-00-00T00:00:00"))
    sorted_offer_id = list(map(lambda v: v.get("couponID", "n/a"), sorted_history))
    len_history = len(sorted_offer_id)
    if len(sorted_history) == 0:
        sorted_history.append({})
    history_processed = {
        "platform": sorted_history[-1].get("platform", "n/a"),
        "dma": sorted_history[-1].get("dma", "n/a"),
        "device_carrier": sorted_history[-1].get("carrier", "n/a"),
        "device_model": sorted_history[-1].get("deviceModel", "n/a"),
        "offer_id": sorted_offer_id[max(0, len_history - 5):] + ["n/a"] * max(0, 5 - len_history),
    }
    return history_processed


def transform_history_to_mxnet_in(history, offerid2idx, platform2idx, dma2idx, carrier2idx, model2idx):
    offer_idx = [offerid2idx.get(offer_id, 0) for offer_id in history["offer_id"]]
    platform_idx = platform2idx.get(history["platform"], 0)  # web is 1
    dma_idx = dma2idx.get(history["dma"], 0)
    device_carrier_idx = carrier2idx.get(history["device_carrier"], 0)
    device_model_idx = model2idx.get(history["device_model"], 0)
    valid_len = max(list(np.nonzero(offer_idx)[0]) + [-1]) + 1

    model_input = mx.io.NDArrayIter(data={'data0': np.asarray([offer_idx]),
                                          'data1': np.asarray([int(valid_len)]),
                                          'data2': np.asarray([int(device_carrier_idx)]),
                                          'data3': np.asarray([int(device_model_idx)]),
                                          'data4': np.asarray([int(dma_idx)]),
                                          'data5': np.asarray([int(platform_idx)])},
                                    batch_size=1)
    return model_input


def transform_mxnet_out_to_offer(model_out, idx2offer):
    topk_idx = model_out.argsort().flatten()[::-1]
    topk_offer = [idx2offer.get(idx, 0) for idx in topk_idx if idx != 0]
    return topk_offer


# Helper Functions - Recommend Rules

def is_offer_type_in(offer_attr_lookup, offer_id, offer_type_list):
    return offer_attr_lookup.get(offer_id, {}).get("Offer Type") in offer_type_list


def is_lapsed(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Offer Type") in ["Lapsed_60", "Lapsed_90"]


def is_offer_type_group_breakfast(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Group Name") == "Breakfast"


def is_offer_type_group_family(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Group Name") == "Family"


def is_offer_type_group_single_serve(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Group Name") == "Single Serve"


def is_offer_type_group_for2(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Group Name") == "For 2"


def is_offer_type_group_snack(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Group Name") == "Snack"


def is_structure_bundle(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Offer Structure") == "Bundle"


def is_structure_sides(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Offer Structure") == "Side/Dessert"


def is_breakfast(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Product Category") == "Breakfast"


def is_discount_low(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Discount Level") == "Low"


def is_discount_med(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Discount Level") == "Med"


def is_discount_high(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Discount Level") == "High"


def is_discount_na(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Discount Level") == "NA"


def get_offer_group(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Personalized Offer Group", "0")


def get_offer_type(offer_attr_lookup, offer_id):
    return offer_attr_lookup.get(offer_id, {}).get("Offer Type")


def get_offer_attributes(offer_attr_lookup, offer_id):
    offer_attributes = {
        "offer_group": get_offer_group(offer_attr_lookup, offer_id),
        # "breakfast": None,
        # "discount_level": None,
        # "offer_structure": None,
        "offer_type_group": None,
        "offer_type": get_offer_type(offer_attr_lookup, offer_id)
    }
    # offer_structure
    if is_offer_type_group_breakfast(offer_attr_lookup, offer_id):
        offer_attributes["offer_type_group"] = "br"
    elif is_offer_type_group_family(offer_attr_lookup, offer_id):
        offer_attributes["offer_type_group"] = "family"
    elif is_offer_type_group_single_serve(offer_attr_lookup, offer_id):
        offer_attributes["offer_type_group"] = "1serve"
    elif is_offer_type_group_for2(offer_attr_lookup, offer_id):
        offer_attributes["offer_type_group"] = "for2"
    elif is_offer_type_group_snack(offer_attr_lookup, offer_id):
        offer_attributes["offer_type_group"] = "snack"
    else:
        offer_attributes["offer_type_group"] = "other"
    # # breakfast
    # if is_breakfast(offer_attr_lookup, offer_id):
    #     offer_attributes["breakfast"] = "br"
    # else:
    #     offer_attributes["breakfast"] = "nbr"
    # # discount_level
    # if is_discount_high(offer_attr_lookup, offer_id):
    #     offer_attributes["discount_level"] = "h"
    # elif is_discount_med(offer_attr_lookup, offer_id):
    #     offer_attributes["discount_level"] = "m"
    # elif is_discount_low(offer_attr_lookup, offer_id):
    #     offer_attributes["discount_level"] = "l"
    # else:
    #     offer_attributes["discount_level"] = "na"
    # # offer_structure
    # if is_structure_bundle(offer_attr_lookup, offer_id):
    #     offer_attributes["offer_structure"] = "bundle"
    # elif is_structure_sides(offer_attr_lookup, offer_id):
    #     offer_attributes["offer_structure"] = "sides"
    # else:
    #     offer_attributes["offer_structure"] = "other"
    return offer_attributes


def update_cluster_rules(topk_offer_attr, cluster_rules, offer_attr_lookup, offer_id, use_cluster_rules=True):
    offer_attributes = get_offer_attributes(offer_attr_lookup, offer_id)
    group = offer_attributes["offer_group"]
    # br_nbr = offer_attributes["breakfast"]
    # h_m_l = offer_attributes["discount_level"]
    # structure = offer_attributes["offer_structure"]
    type_group = offer_attributes["offer_type_group"]
    offer_type = offer_attributes["offer_type"]
    topk_offer_attr = topk_offer_attr.copy()
    if use_cluster_rules:
        # satisfy_rule = (topk_offer_attr.get(br_nbr, 0) < cluster_rules.get(br_nbr, 999)
        #                 and topk_offer_attr.get(h_m_l, 0) < cluster_rules.get(h_m_l, -1)
        #                 and (group == "0" or group not in topk_offer_attr["groups"]))
        satisfy_rule = (
                (topk_offer_attr.get(type_group, 0) < cluster_rules.get(type_group, 999))
                and (group == "0" or group not in topk_offer_attr["groups"])
        )
    else:
        satisfy_rule = (group == "0" or group not in topk_offer_attr["groups"])
    if satisfy_rule:
        # topk_offer_attr[br_nbr] += 1
        # topk_offer_attr[h_m_l] += 1
        # topk_offer_attr[structure] += 1
        topk_offer_attr[type_group] += 1
        topk_offer_attr["groups"].append(group)
        topk_offer_attr[offer_type] += 1
    return satisfy_rule, topk_offer_attr


def transform_offer_attr(offer_attr):
    offer_attr_log = (
            "br-{},family-{},for2-{},1serve-{},snack-{},other-{}|".format(offer_attr["br"], offer_attr["family"],
                                                                          offer_attr["for2"], offer_attr["1serve"],
                                                                          offer_attr["snack"], offer_attr["other"])
            + "P1-{},P2-{},L60-{},L90-{}|".format(offer_attr["1_Personalized"], offer_attr["2_Personalized"],
                                                  offer_attr["Lapsed_60"], offer_attr["Lapsed_90"])
            + "groups-{}".format(str(offer_attr["groups"]))
    )
    # offer_attr_log = (
    #     "br-{},nbr-{}|".format(offer_attr["br"], offer_attr["nbr"])
    #     + "h-{},m-{},l-{},na-{}|".format(offer_attr["h"], offer_attr["m"], offer_attr["l"], offer_attr["na"])
    #     + "bundle-{},sides-{},other-{}|".format(offer_attr["bundle"], offer_attr["sides"], offer_attr["other"])
    #     + "P1-{},P2-{},L60-{},L90-{}|".format(offer_attr["1_Personalized"], offer_attr["2_Personalized"],
    #                                           offer_attr["Lapsed_60"], offer_attr["Lapsed_90"])
    #     + "groups-{}".format(str(offer_attr["groups"]))
    # )
    return offer_attr_log


# Main functions

def load_model(model_name, model_prefix, model_epoch, context_lookup_names, sequence_lookup_name):
    # lookup tables
    model_path = os.path.join("models/", model_name, model_prefix)
    lookup_paths = {
        k: os.path.join("models/", model_name, k + "_lookup.csv") for k in context_lookup_names + [sequence_lookup_name]
    }
    lookup_tables = {
        k: pd.read_csv(lookup_paths[k]) for k in context_lookup_names + [sequence_lookup_name]
    }
    context_to_idx = {
        k: generate_lookup_dict(lookup_tables[k], key_col=k, value_col=k + "_idx") for k in context_lookup_names
    }
    sequence_to_idx = generate_lookup_dict(df=lookup_tables[sequence_lookup_name],
                                           key_col=sequence_lookup_name + "_id",
                                           value_col=sequence_lookup_name + "_idx")
    idx_to_sequence = generate_lookup_dict(df=lookup_tables[sequence_lookup_name],
                                           key_col=sequence_lookup_name + "_idx",
                                           value_col=sequence_lookup_name + "_id")

    # model
    model = mx.model.FeedForward.load(prefix=model_path, epoch=model_epoch, ctx=mx.cpu())
    return model, (context_to_idx, sequence_to_idx, idx_to_sequence),


def load_offer_metadata(offer_pool_path, s3_id_mapping_bucket, s3_id_mapping_prefix, s3_rest_tz_bucket,
                        s3_rest_tz_prefix):
    # id_mapping
    keys = []
    id_mapping_dfs = []
    s3 = boto3.client('s3')
    list_obj = s3.list_objects_v2(Bucket=s3_id_mapping_bucket, Prefix=s3_id_mapping_prefix)
    assert list_obj["KeyCount"] < 1000
    for content in list_obj["Contents"]:
        if content["Key"].endswith(".csv"):
            keys.append(content["Key"])
    assert len(keys) == 1
    print("Load %d csv files from %s" % (len(keys), os.path.dirname(keys[0])))
    for i, key in enumerate(keys):
        obj = s3.get_object(Bucket=s3_id_mapping_bucket, Key=key)
        id_mapping_dfs.append(pd.read_csv(obj["Body"], dtype="str"))
    offer_id_mapping_df = pd.concat(id_mapping_dfs)
    active_offer_pool = offer_id_mapping_df[offer_id_mapping_df["activeOffer"] == "Y"]["id"].tolist()
    offer_id_mapping = offer_id_mapping_df[
        ["id", "name", "localizedImageUrl"]].fillna("n/a").set_index(keys="id").to_dict(orient="index")

    # offer pool and offer attributes
    offer_pools, offer_attributes = load_offer_id_from_pool(offer_pool_path, active_offer_pool)

    # rest_tz
    keys = []
    rest_tz_dfs = []
    s3 = boto3.client('s3')
    list_obj = s3.list_objects_v2(Bucket=s3_rest_tz_bucket, Prefix=s3_rest_tz_prefix)
    assert list_obj["KeyCount"] < 1000
    for content in list_obj["Contents"]:
        if content["Key"].endswith(".csv"):
            keys.append(content["Key"])
    assert len(keys) == 1
    print("Load %d csv files from %s" % (len(keys), os.path.dirname(keys[0])))
    for i, key in enumerate(keys):
        obj = s3.get_object(Bucket=s3_rest_tz_bucket, Key=key)
        rest_tz_dfs.append(pd.read_csv(obj["Body"], dtype="str"))
    rest_tz_df = pd.concat(rest_tz_dfs)

    return (offer_pools, offer_attributes), (active_offer_pool, offer_id_mapping), rest_tz_df


def mxnet_predict(model, history_raw, offer2idx, idx2offer, carrier2idx, dma2idx, model2idx, platform2idx):
    """
    API post -> raw -> idx_in -> model -> idx_out -> raw -> API
    """
    history_processed = transform_raw_to_history(history_raw=history_raw)
    model_in = transform_history_to_mxnet_in(history=history_processed,
                                             offerid2idx=offer2idx,
                                             platform2idx=platform2idx,
                                             dma2idx=dma2idx,
                                             carrier2idx=carrier2idx,
                                             model2idx=model2idx)
    model_out = model.predict(model_in)
    recommended_offer_id = transform_mxnet_out_to_offer(model_out=model_out, idx2offer=idx2offer)

    return recommended_offer_id


def filter_raw_by_rules(offer_attr_lookup, weekly_pool, sorted_offer_id, national_adhoc_offers, price_tier, lapsed,
                        topk, lapsed_rule, cluster_rules, price_tier_mapper, top_performance_offers):
    """
    Filter sorted model_out by rules generated by user cluster
    :param offer_attr_lookup: offer_attr_lookup table containing offer info
    :param weekly_pool: list of weekly SENDABLE offer_id
    :param sorted_offer_id: list of sorted offer_id, descending
    :param national_adhoc_offers: list of national offer_id that has already been assigned
    :param price_tier: str "1_Personalized" (high) or "2_Personalized" (low)
    :param lapsed: str "0", "30", "60" or "90"
    :param topk: number of offers returned
    :param lapsed_rule: number of lapsed offer assigned to lapsed user
    :param cluster_rules:
    :param price_tier_mapper: {"1to2": {oid1: oid2}, "2to1": {oid2: oid1}}
    :param top_performance_offers:
    :return: list of sorted and filtered offer_id that satisfies all given rules
    """
    topk_offer_id = []
    topk_offer_attr = {
        # "br": 0,
        # "nbr": 0,
        # "h": 0,
        # "m": 0,
        # "l": 0,
        # "na": 0,
        # "bundle": 0,
        # "sides": 0,
        "br": 0,
        "family": 0,
        "1serve": 0,
        "for2": 0,
        "snack": 0,
        "other": 0,
        "1_Personalized": 0,
        "2_Personalized": 0,
        "Lapsed_60": 0,
        "Lapsed_90": 0,
        "groups": [],
    }

    # TODO: remove this in future
    if price_tier == "1_Personalized":
        # manually select personalized offers from top performance offers - 4 out of 5
        manual_assigned = [
            top_performance_offers["tier1"]["br"][0],
            top_performance_offers["tier1"]["family"][0],
            top_performance_offers["tier1"]["for2"][0],
            top_performance_offers["tier1"]["1serve"][0],
        ]
        topk_offer_id += manual_assigned
        # TODO: update using update_cluster_rule
        # update topk_offer_attr
        topk_offer_attr["br"] = 1
        topk_offer_attr["family"] = 1
        topk_offer_attr["for2"] = 1
        topk_offer_attr["1serve"] = 1
        topk_offer_attr["1_Personalized"] = 4
    else:
        # manually select personalized offers from top performance offers - 4 out of 5
        manual_assigned = [
            top_performance_offers["tier2"]["family"][0],
            top_performance_offers["tier2"]["for2"][0],
            top_performance_offers["tier2"]["1serve"][0],
            top_performance_offers["tier2"]["1serve"][1],
        ]
        topk_offer_id += manual_assigned
        # TODO: update using update_cluster_rule
        # update topk_offer_attr
        topk_offer_attr["family"] = 1
        topk_offer_attr["for2"] = 1
        topk_offer_attr["1serve"] = 2
        topk_offer_attr["2_Personalized"] = 4
    mandatory_offers_actual = national_adhoc_offers.copy() + topk_offer_id.copy()
    mandatory_offer_groups = [
        get_offer_group(offer_attr_lookup, oid)
        for oid in mandatory_offers_actual
    ]
    topk_offer_attr["groups"] = mandatory_offer_groups

    # filter SENDABLE
    if lapsed == "90":
        offer_type_list = [price_tier, "Lapsed_90", "Lapsed_60"]
    elif lapsed == "60":
        offer_type_list = [price_tier, "Lapsed_60"]
    else:
        offer_type_list = [price_tier]
    # TODO: add adhoc: let tier1 offers flow to tier2 users - DONE
    if price_tier == "2_Personalized":
        offer_type_list.append("1_Personalized")
    ####
    if price_tier == "1_Personalized":
        price_tier_mapping = price_tier_mapper["2to1"]
    elif price_tier == "2_Personalized":
        price_tier_mapping = price_tier_mapper["1to2"]
    else:
        price_tier_mapping = price_tier_mapper["1to2"]
    offer_id_sendable_sorted = [
        price_tier_mapping.get(oid, oid)
        for oid in sorted_offer_id
    ]
    offer_id_sendable_sorted = [
        oid
        for oid in offer_id_sendable_sorted
        if oid in weekly_pool
        and oid not in mandatory_offers_actual
        and oid not in topk_offer_id
        and is_offer_type_in(offer_attr_lookup, oid, offer_type_list)
    ]

    # filter LAPSED
    if str(lapsed) == "90":
        lapsed_count = 0
        for offer_id in offer_id_sendable_sorted:
            if is_offer_type_in(offer_attr_lookup, offer_id, ["Lapsed_90", "Lapsed_60"]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules=cluster_rules,
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "lapsed", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    lapsed_count += 1
                    if lapsed_count >= lapsed_rule or len(topk_offer_id) >= topk:
                        break
    elif str(lapsed) == "60":
        lapsed_count = 0
        for offer_id in offer_id_sendable_sorted:
            if is_offer_type_in(offer_attr_lookup, offer_id, ["Lapsed_60"]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules=cluster_rules,
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "lapsed", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    lapsed_count += 1
                    if lapsed_count >= lapsed_rule or len(topk_offer_id) >= topk:
                        break

    # # filter base rules: 1 bundle and 1 sides
    # offer_id_sendable_sorted = list(filter(
    #     lambda x: (x not in topk_offer_id) and (not is_lapsed(offer_attr_lookup, x)),
    #     offer_id_sendable_sorted
    # ))
    # if len(topk_offer_id) < topk and topk_offer_attr["bundle"] < 1:
    #     for offer_id in offer_id_sendable_sorted:
    #         if is_structure_bundle(offer_attr_lookup, offer_id):
    #             satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
    #                                                                  cluster_rules=cluster_rules,
    #                                                                  offer_attr_lookup=offer_attr_lookup,
    #                                                                  offer_id=offer_id,
    #                                                                  use_cluster_rules=True)
    #             if satisfy_rule:
    #                 # print(offer_id, "bundle", topk_offer_attr)
    #                 topk_offer_id.append(offer_id)
    #                 break
    # offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
    # if len(topk_offer_id) < topk and topk_offer_attr["sides"] < 1:
    #     for offer_id in offer_id_sendable_sorted:
    #         if is_structure_sides(offer_attr_lookup, offer_id):
    #             satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
    #                                                                  cluster_rules=cluster_rules,
    #                                                                  offer_attr_lookup=offer_attr_lookup,
    #                                                                  offer_id=offer_id,
    #                                                                  use_cluster_rules=True)
    #             if satisfy_rule:
    #                 # print(offer_id, "sides", topk_offer_attr)
    #                 topk_offer_id.append(offer_id)
    #                 break

    # filter cluster rules
    offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
    if len(topk_offer_id) < topk:
        for offer_id in offer_id_sendable_sorted:
            if is_offer_type_in(offer_attr_lookup, offer_id, [price_tier]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules=cluster_rules,
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "cluster", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    if len(topk_offer_id) >= topk:
                        break

    # fill rest of the array
    if len(topk_offer_id) < topk:
        # print("drop cluster rules")
        offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
        for offer_id in offer_id_sendable_sorted:
            if is_offer_type_in(offer_attr_lookup, offer_id, [price_tier]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules=cluster_rules,
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=False)
                if satisfy_rule:
                    # print(offer_id, "cluster", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    if len(topk_offer_id) >= topk:
                        break

    return topk_offer_id, transform_offer_attr(topk_offer_attr)


def filter_randomized(offer_attr_lookup, sorted_offer_id, national_adhoc_offers, price_tier, lapsed, lapsed_rule, topk):
    topk_offer_id = []
    topk_offer_attr = {
        "br": 0,
        "family": 0,
        "1serve": 0,
        "for2": 0,
        "snack": 0,
        "other": 0,
        "1_Personalized": 0,
        "2_Personalized": 0,
        "Lapsed_60": 0,
        "Lapsed_90": 0,
        "groups": [],
    }
    national_adhoc_offers_groups = [
        get_offer_group(offer_attr_lookup, oid)
        for oid in national_adhoc_offers
    ]
    topk_offer_attr["groups"] = national_adhoc_offers_groups

    # filter SENDABLE
    if lapsed == "90" or lapsed == "60":
        offer_type_list = [price_tier, "Lapsed_90", "Lapsed_60"]
    elif lapsed == "60":
        offer_type_list = [price_tier, "Lapsed_60"]
    else:
        offer_type_list = [price_tier]
    offer_id_sendable = [
        oid
        for oid in sorted_offer_id
        if oid not in national_adhoc_offers
        and is_offer_type_in(offer_attr_lookup, oid, offer_type_list)
    ]

    if str(lapsed) == "90":
        lapsed_count = 0
        for offer_id in offer_id_sendable:
            if is_offer_type_in(offer_attr_lookup, offer_id, ["Lapsed_90", "Lapsed_60"]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules={},
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=False)
                if satisfy_rule:
                    # print(offer_id, "lapsed", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    lapsed_count += 1
                    if lapsed_count >= lapsed_rule or len(topk_offer_id) >= topk:
                        break
    elif str(lapsed) == "60":
        lapsed_count = 0
        for offer_id in offer_id_sendable:
            if is_offer_type_in(offer_attr_lookup, offer_id, ["Lapsed_60"]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules={},
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=False)
                if satisfy_rule:
                    # print(offer_id, "lapsed", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    lapsed_count += 1
                    if lapsed_count >= lapsed_rule or len(topk_offer_id) >= topk:
                        break

    # fill rest of the array
    if len(topk_offer_id) < topk:
        for offer_id in offer_id_sendable:
            if is_offer_type_in(offer_attr_lookup, offer_id, [price_tier]):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr=topk_offer_attr,
                                                                     cluster_rules={},
                                                                     offer_attr_lookup=offer_attr_lookup,
                                                                     offer_id=offer_id,
                                                                     use_cluster_rules=False)
                if satisfy_rule:
                    # print(offer_id, "cluster", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    if len(topk_offer_id) >= topk:
                        break

    return topk_offer_id, transform_offer_attr(topk_offer_attr)


def get_dma(store_ids, user_location, dim_rest):
    """
    :param store_ids: list of store_id strings
    :param user_location: user location parsed from dynamodb
    :param dim_rest: pandas dataframe of rest_tz_bk_full
    :return: string of the most frequent dma
    """
    tier_1_dma = [
        {'name': 'AK', 'tlog': 29, 'geoip': 745},
        {'name': 'HI', 'tlog': 340, 'geoip': 744},
        {'name': 'Chico-Redding', 'tlog': 148, 'geoip': 868},
        {'name': 'Fresno-Visalia', 'tlog': 284, 'geoip': 866},
        {'name': 'Hartford-New Haven', 'tlog': 332, 'geoip': 533},
        {'name': 'Las Vegaas', 'tlog': 420, 'geoip': 839},
        {'name': 'Long Island', 'tlog': 531, 'geoip': 501},
        {'name': 'Los Angeles', 'tlog': 444, 'geoip': 803},
        {'name': 'Medford-Klamath Falls', 'tlog': 476, 'geoip': 813},
        {'name': 'New Jersey - Sub', 'tlog': 530, 'geoip': 501},
        {'name': 'NYC Sub', 'tlog': 529, 'geoip': 501},
        {'name': 'Phoenix', 'tlog': 584, 'geoip': 753},
        {'name': 'San Diego', 'tlog': 684, 'geoip': 825},
        {'name': 'San Francisco-Oak-San Jose', 'tlog': 688, 'geoip': 807},
        {'name': 'Santa Barbara', 'tlog': 696, 'geoip': 855},
        {'name': 'Springfield-Holyoke', 'tlog': 728, 'geoip': 543},
        {'name': 'Tucson-Nogales', 'tlog': 772, 'geoip': 789},
        {'name': 'Providence-New Beford', 'tlog': 608, 'geoip': 521},
        {'name': 'Palm Springs', 'tlog': 564, 'geoip': 804},
        {'name': 'Boston', 'tlog': 96, 'geoip': 807},
    ]
    lookup = dict(zip(dim_rest.rest_no.map(str), dim_rest.dma_no))
    dma_nos = [lookup.get(_) for _ in store_ids if _ in lookup]
    if dma_nos:
        user_dma = max(set(dma_nos), key=dma_nos.count)
        user_tier = "1" if user_dma in [str(dma.get("tlog")) for dma in tier_1_dma] else "2"
        is_phoenix = True if user_dma in ["584", "772"] else False
    elif user_location:
        user_dma = str(user_location.get('dmaCode'))
        user_tier = "1" if user_dma in [str(dma.get("geoip")) for dma in tier_1_dma] else "2"
        is_phoenix = True if user_dma in ["753", "789"] else False
    else:
        # user_dma = "n/a"
        user_tier = "2"
        is_phoenix = False
    return user_tier, is_phoenix



import torch
from torch import nn
from torch.nn import TransformerEncoder, TransformerEncoderLayer
from .txt import MeanMaxPooling, PositionalEncoding, ContextTransformer, SequenceTransformerHistory
from .mmoe import MMoE


# Transformer x Embedding x Embedding as Bottom

class SequenceTransformerHistoryLite(SequenceTransformerHistory):
    def __init__(self, seq_num, cross_size, seq_embed_size=200, seq_max_length=8, seq_num_heads=4,
                 seq_hidden_size=512, seq_transformer_dropout=0.0, seq_num_layers=2, seq_pooling_dropout=0.0,
                 seq_pe=True):
        super(SequenceTransformerHistory, self).__init__()
        self.seq_embedding = nn.Embedding(seq_num, seq_embed_size)
        self.seq_pos = seq_pe
        self.seq_embed_size = seq_embed_size
        if seq_pe:
            self.pos_encoder = PositionalEncoding(d_model=seq_embed_size,
                                                  dropout=seq_transformer_dropout,
                                                  max_len=seq_max_length)
        encoder_layers = TransformerEncoderLayer(d_model=seq_embed_size,
                                                 nhead=seq_num_heads,
                                                 dropout=seq_transformer_dropout,
                                                 dim_feedforward=seq_hidden_size,
                                                 activation='relu',
                                                 batch_first=True)
        self.seq_encoder = TransformerEncoder(encoder_layers, num_layers=seq_num_layers)
        self.seq_pooling_dp = MeanMaxPooling(dropout=seq_pooling_dropout)
        self.seq_dense = torch.nn.Linear(in_features=2 * seq_embed_size, out_features=cross_size)


class ContextEmbedding(nn.Module):
    """
    [[B, ] * C] -> [B, cross_size]
    """
    def __init__(self, ctx_nums, cross_size, ctx_embed_size=100, ctx_num_heads=2,
                 ctx_hidden_size=512, ctx_num_layers=1, ctx_transformer_dropout=0.0,
                 ctx_pooling_dropout=0.0, ctx_pe=False):
        super().__init__()
        if isinstance(ctx_embed_size, int):
            self.ctx_embedding = nn.ModuleList([
                nn.Embedding(ctx_num, ctx_embed_size)
                for ctx_num in ctx_nums
            ])
            dense_in = len(ctx_nums) * ctx_embed_size
        elif isinstance(ctx_embed_size, list) or isinstance(ctx_embed_size, tuple):
            self.ctx_embedding = nn.ModuleList([
                nn.Embedding(ctx_num, ctx_embed)
                for ctx_num, ctx_embed in zip(ctx_nums, ctx_embed_size)
            ])
            dense_in = sum(ctx_embed_size)
        else:
            raise NotImplementedError()
        self.ctx_pe = False
        self.ctx_embed_size = ctx_embed_size
        self.ctx_dense = torch.nn.Linear(in_features=dense_in, out_features=cross_size)

    def forward(self, ctx_in):
        """
        :param ctx_in: list, a list of Tensor of shape [batch_size, 1]
        :return: Tensor, shape [batch_size, cross_size]
        """
        # [[B, ] * C]
        ctx_embedding_list = [self.ctx_embedding[i](input_ctx).unsqueeze(1)
                              for i, input_ctx in enumerate(ctx_in)]  # -> [(B, 1, E_i) * C]
        ctx_out = torch.cat(ctx_embedding_list, dim=2).squeeze(1)  # -> [B, sum(E_i)]
        ctx_out = self.ctx_dense(ctx_out)  # -> (B, cross_size)
        return ctx_out


class TxEBottom(nn.Module):
    def __init__(self, ctx_nums, seq_num, ctx_cross_size=100, seq_cross_size=200, is_candidate_mode=True,
                 context_transformer_kwargs=None, sequence_transformer_kwargs=None):
        super().__init__()
        context_transformer_kwargs = context_transformer_kwargs if context_transformer_kwargs else {}
        sequence_transformer_kwargs = sequence_transformer_kwargs if sequence_transformer_kwargs else {}
        self.is_candidate_mode = is_candidate_mode
        self.features_dim = seq_cross_size + ctx_cross_size
        self.context_transformer = ContextEmbedding(
            ctx_nums=ctx_nums,
            cross_size=seq_cross_size,
            **context_transformer_kwargs,
        )
        self.sequence_transformer = SequenceTransformerHistoryLite(
            seq_num=seq_num,
            cross_size=ctx_cross_size,
            **sequence_transformer_kwargs,
        )
        if is_candidate_mode:
            # self.candidate_dense = nn.Linear(
            #     in_features=self.sequence_transformer.seq_embed_size,
            #     out_features=cross_size
            # )
            pass

    def forward(self, ctx_in, seq_in, vl_in, candidate_in, seq_history=None):
        """
        :param ctx_in: list, a list of Tensor of shape [batch_size, 1]
        :param seq_in: Tensor, shape [batch_size, seq_len]
        :param vl_in: Tensor, shape [batch_size]
        :param candidate_in: Tensor, shape [batch_size]
        :param seq_history: Tensor, shape [batch_size, history_len]
        :return:
        """
        # input [[B, 1] * C] and [B, 5]
        ctx_out = self.context_transformer(ctx_in=ctx_in)
        seq_out = self.sequence_transformer(seq_in=seq_in, vl_in=vl_in, seq_history=seq_history)
        outs = torch.cat([seq_out, ctx_out], dim=1)  # -> [B, CROSS_seq + CROSS_ctx]
        if self.is_candidate_mode:
            candidate_embed = self.sequence_transformer.seq_embedding(candidate_in)
            outs = torch.concat([outs, candidate_embed], dim=1)  # -> [B, CROSS_seq + CROSS_ctx + seq_embed_size]
        return outs


class MultiTaskTxE(nn.Module):
    def __init__(self, ctx_nums, seq_num, expert_num, expert_hidden_sizes,
                 task_num, task_hidden_sizes, task_last_activations,
                 ctx_cross_size=100, seq_cross_size=200, is_candidate_mode=True,
                 context_transformer_kwargs=None, sequence_transformer_kwargs=None):
        super().__init__()
        self.is_candidate_mode = is_candidate_mode
        self.shared_bottom = TxEBottom(
            ctx_nums=ctx_nums,
            seq_num=seq_num,
            ctx_cross_size=ctx_cross_size,
            seq_cross_size=seq_cross_size,
            is_candidate_mode=is_candidate_mode,
            context_transformer_kwargs=context_transformer_kwargs,
            sequence_transformer_kwargs=sequence_transformer_kwargs,
        )
        mmoe_input_size = ctx_cross_size + seq_cross_size + self.shared_bottom.sequence_transformer.seq_embed_size
        self.mmoe = MMoE(
            input_size=mmoe_input_size,
            expert_num=expert_num,
            expert_hidden_sizes=expert_hidden_sizes,
            task_num=task_num,
            task_hidden_sizes=task_hidden_sizes,
            task_last_activations=task_last_activations,
        )

    def forward(self, ctx_in, seq_in, vl_in, candidate_in, seq_history=None):
        bottom_features = self.shared_bottom(ctx_in, seq_in, vl_in, candidate_in, seq_history)
        outs = self.mmoe(bottom_features)
        return outs


# Transformer x Embedding x Embedding as Bottom

class TxTBottom(nn.Module):
    def __init__(self, ctx_nums, seq_num, cross_size=200, is_candidate_mode=True,
                 context_transformer_kwargs=None, sequence_transformer_kwargs=None):
        super().__init__()
        context_transformer_kwargs = context_transformer_kwargs if context_transformer_kwargs else {}
        sequence_transformer_kwargs = sequence_transformer_kwargs if sequence_transformer_kwargs else {}
        self.is_candidate_mode = is_candidate_mode
        self.features_dim = cross_size
        self.context_transformer = ContextTransformer(
            ctx_nums=ctx_nums,
            cross_size=cross_size,
            **context_transformer_kwargs,
        )
        self.sequence_transformer = SequenceTransformerHistory(
            seq_num=seq_num,
            cross_size=cross_size,
            **sequence_transformer_kwargs,
        )
        if is_candidate_mode:
            # self.candidate_dense = nn.Linear(
            #     in_features=self.sequence_transformer.seq_embed_size,
            #     out_features=cross_size
            # )
            pass

    def forward(self, ctx_in, seq_in, vl_in, candidate_in, seq_history=None):
        """
        :param ctx_in: list, a list of Tensor of shape [batch_size, 1]
        :param seq_in: Tensor, shape [batch_size, seq_len]
        :param vl_in: Tensor, shape [batch_size]
        :param candidate_in: Tensor, shape [batch_size]
        :param seq_history: Tensor, shape [batch_size, history_len]
        :return:
        """
        # input [[B, 1] * C] and [B, 5]
        ctx_out = self.context_transformer(ctx_in=ctx_in)
        seq_out = self.sequence_transformer(seq_in=seq_in, vl_in=vl_in, seq_history=seq_history)
        outs = torch.mul(seq_out, ctx_out)  # -> [B, cross_size]
        if self.is_candidate_mode:
            candidate_embed = self.sequence_transformer.seq_embedding(candidate_in)
            outs = torch.concat([outs, candidate_embed], dim=1)  # -> [B, seq_embed_size]
        return outs


class MultiTaskTxT(nn.Module):
    def __init__(self, ctx_nums, seq_num, expert_num, expert_hidden_sizes,
                 task_num, task_hidden_sizes, task_last_activations,
                 cross_size=200, is_candidate_mode=True,
                 context_transformer_kwargs=None, sequence_transformer_kwargs=None):
        super().__init__()
        self.is_candidate_mode = is_candidate_mode
        self.shared_bottom = TxTBottom(
            ctx_nums=ctx_nums,
            seq_num=seq_num,
            cross_size=cross_size,
            is_candidate_mode=is_candidate_mode,
            context_transformer_kwargs=context_transformer_kwargs,
            sequence_transformer_kwargs=sequence_transformer_kwargs,
        )
        mmoe_input_size = cross_size + self.shared_bottom.sequence_transformer.seq_embed_size
        self.mmoe = MMoE(
            input_size=mmoe_input_size,
            expert_num=expert_num,
            expert_hidden_sizes=expert_hidden_sizes,
            task_num=task_num,
            task_hidden_sizes=task_hidden_sizes,
            task_last_activations=task_last_activations,
        )

    def forward(self, ctx_in, seq_in, vl_in, candidate_in, seq_history=None):
        bottom_features = self.shared_bottom(ctx_in, seq_in, vl_in, candidate_in, seq_history)
        outs = self.mmoe(bottom_features)
        return outs



import logging
import pandas as pd
import mxnet as mx
from mxnet.gluon.data import Dataset
from mxnet.gluon.contrib.estimator.batch_processor import BatchProcessor
from mxnet import ndarray, autograd
from mxnet import numpy as _mx_np
from mxnet.util import is_np_array
import boto3


class DownloadS3(object):
    def __init__(self, bucket, train_prefix, eval_prefix, parts_num=9999):
        self.bucket = bucket
        self.train_prefix = train_prefix
        self.eval_prefix = eval_prefix
        self.parts_num = parts_num
        self.data = None

    def download_train(self):
        keys = []
        df_list = []
        s3_client = boto3.client('s3', verify=False)

        resp = s3_client.list_objects_v2(Bucket=self.bucket, Prefix=self.train_prefix)
        for obj in resp['Contents']:
            if obj['Key'].endswith(".json"):
                keys.append(obj['Key'])

        for i, key in enumerate(keys):
            obj = s3_client.get_object(Bucket=self.bucket, Key=key)
            df = pd.read_json(obj['Body'], orient='columns', lines=True)
            df_list.append(df)
            if i > self.parts_num:
                break

        return pd.concat(df_list)

    def download_test(self):
        keys = []
        df_list = []
        s3_client = boto3.client('s3', verify=False)

        resp = s3_client.list_objects_v2(Bucket=self.bucket, Prefix=self.eval_prefix)
        for obj in resp['Contents']:
            if obj['Key'].endswith(".json"):
                keys.append(obj['Key'])

        for i, key in enumerate(keys):
            obj = s3_client.get_object(Bucket=self.bucket, Key=key)
            df = pd.read_json(obj['Body'], orient='columns', lines=True)
            df_list.append(df)
            if i > self.parts_num:
                break

        return pd.concat(df_list)


class DataIterLoaderTxT:
    def __init__(self, data_iter, seq_col, vl_col, context_cols):
        self.data_iter = data_iter
        self.seq_col = seq_col
        self.vl_col = vl_col
        self.context_cols = context_cols

    def __iter__(self):
        self.data_iter.reset()
        return self

    def __next__(self):
        batch = self.data_iter.__next__()
        data = batch.data
        label = batch.label[0]
        desc_list = list(x[0] for x in self.data_iter.provide_data)
        sequence_idx = desc_list.index(self.seq_col)
        vl_idx = desc_list.index(self.vl_col)
        context_idx = [desc_list.index(c) for c in self.context_cols]
        sequence_data = data[sequence_idx]
        valid_length = data[vl_idx]
        context_data = [data[i] for i in context_idx]
        return (sequence_data, valid_length, context_data), label

    def next(self):
        return self.__next__()


class BatchProcessorTxT(BatchProcessor):
    def __init__(self):
        pass

    def _get_data_and_label(self, batch, ctx, batch_axis=0):
        assert len(ctx) == 1, "Multiple CTX is not supported"
        array_fn = _mx_np.array if is_np_array() else ndarray.array

        data = [[array_fn(d, ctx[0]).as_in_context(ctx[0]) for d in batch[0]]]
        label = [array_fn(batch[1], ctx[0]).as_in_context(ctx[0])]
        # data = split_and_load(data, ctx_list=ctx, batch_axis=batch_axis)
        # label = split_and_load(label, ctx_list=ctx, batch_axis=batch_axis)
        return data, label

    def evaluate_batch(self, estimator,
                       val_batch,
                       batch_axis=0):
        data, label = self._get_data_and_label(val_batch, estimator.context, batch_axis)

        pred = [estimator.val_net(seq, vl, contexts, None) for seq, vl, contexts in data]
        loss = [estimator.val_loss(y_hat, y) for y_hat, y in zip(pred, label)]

        return data, label, pred, loss

    def fit_batch(self, estimator,
                  train_batch,
                  batch_axis=0):
        data, label = self._get_data_and_label(train_batch, estimator.context, batch_axis)

        with autograd.record():
            pred = [estimator.net(seq, vl, contexts, None) for seq, vl, contexts in data]
            loss = [estimator.loss(y_hat, y) for y_hat, y in zip(pred, label)]

        for l in loss:
            l.backward()

        return data, label, pred, loss


class DatasetTxT(Dataset):
    def __init__(self, data, seq_col, context_cols, vl_col, label_col):
        self.data = data
        self.seq_col = seq_col
        self.context_cols = context_cols
        self.vl_col = vl_col
        self.label_col = label_col

    def __getitem__(self, idx):
        seq = self.data.iloc[idx][self.seq_col]
        vl = self.data.iloc[idx][self.vl_col]
        contexts = self.data.iloc[idx][self.context_cols].tolist()
        label = self.data.iloc[idx][self.label_col]
        return (seq, vl, contexts), label

    def __len__(self):
        return len(self.data)


def evaluate_accuracy(data_iterator, model, ctx):
    acc = mx.metric.Accuracy()
    acc3 = mx.metric.TopKAccuracy(top_k=3)
    for i, ((sequence_data, valid_length, context_data), label) in enumerate(data_iterator):
        sequence = sequence_data.as_in_context(ctx)
        valid_length = valid_length.as_in_context(ctx)
        context = [d.as_in_context(ctx) for d in context_data]
        label = label.as_in_context(ctx)
        output = model(sequence, valid_length, context, None)
        # predictions = mx.nd.argmax(output, axis=1)
        # predictions3 = mx.nd.slice_axis(mx.nd.argsort(output, axis=1, is_ascend=False), axis=1, begin=0, end=3)
        acc.update(preds=output, labels=label)
        acc3.update(preds=output, labels=label)
    return float(acc.get()[1]), float(acc3.get()[1])


def load_s3_lookup(s3_bucket, s3_lookup, f_format="csv"):
    logger = logging.getLogger("load_s3")
    keys = []
    s3 = boto3.client('s3')
    list_obj = s3.list_objects_v2(Bucket=s3_bucket, Prefix=s3_lookup)
    if list_obj["KeyCount"] > 1000:
        logger.error("Found more than 1000 keys and only read partial data")
    for content in list_obj["Contents"]:
        if content["Key"].endswith(f_format):
            keys.append(content["Key"])
    if len(keys) > 1:
        logger.error("Found more than 1 lookup %s" % f_format)
    obj = s3.get_object(Bucket=s3_bucket, Key=keys[0])
    logger.info("Load lookup table from %s" % keys[0])
    if f_format == "json":
        df = pd.read_json(obj["Body"], lines=True)
    elif f_format == "csv":
        df = pd.read_csv(obj["Body"])
    else:
        raise NotImplementedError()
    return df


def load_s3_offer_embed(s3_bucket, s3_key, mask_embed="zeros", save_local_path=None):
    s3 = boto3.resource('s3')
    obj = s3.Object(s3_bucket, s3_key)
    offer_embed_df = pd.read_json(obj.get()['Body'], lines=True)
    if save_local_path:
        offer_embed_df.to_json(save_local_path, orient="records", lines=True)
    if mask_embed == "zeros":
        embedding_weight = ndarray.array(
            pd.Series([[0] * 100], index=[-1]).append(offer_embed_df["embed_pca"]).reset_index(drop=True).tolist())
        return embedding_weight
    else:
        raise NotImplementedError()




import copy
import os
import shutil
import time
import math
import json
import numpy as np
import pandas as pd
import boto3
import torch
import wandb
from torch.utils.data import Dataset, IterableDataset, DataLoader
import mlflow
import pytorch_lightning as pl
from pytorch_lightning.loggers import WandbLogger
from .callbacks_base import AttnVisCallbackBase, MLflowCallbackBase, TensorboardCallbackBase
from ..models.multitask import MultiTaskTxT, MultiTaskTxE
from pytorch_lightning import LightningModule


class MTWandbLogger(WandbLogger):
    def __init__(self, main_logger, task1_logger, task2_logger, loss_logger):
        super().__init__()
        self.main_logger = main_logger
        self.task1_logger = task1_logger
        self.task2_logger = task2_logger
        self.loss_logger = loss_logger


class MTMLflowCallback(MLflowCallbackBase):
    def update_mlflow_metrics(self, trainer, pl_module, phase):
        trainer_logger = trainer.logger
        trainer_loss_logger = trainer_logger.loss_logger
        trainer_task1_logger = trainer_logger.task1_logger
        trainer_task2_logger = trainer_logger.task2_logger
        mlflow.log_metrics(
            metrics={f"{phase}/loss": trainer_loss_logger.get("loss"),
                     f"{phase}/loss1": trainer_loss_logger.get("loss1"),
                     f"{phase}/loss2": trainer_loss_logger.get("loss2"),
                     f"{phase}/acc": trainer_task1_logger.get("acc"),
                     f"{phase}/f1": trainer_task1_logger.get("f1"),
                     f"{phase}/mae": trainer_task2_logger.get("mae"),
                     f"{phase}/mse": trainer_task2_logger.get("mse")},
            step=pl_module.current_epoch
        )


class MTTopKAccCallback(pl.callbacks.Callback):
    def __init__(self, test_dataset, top_k, device):
        self.top_k = top_k
        self.device = device
        test_dataset = copy.deepcopy((test_dataset))
        # remove negative sampled data points
        pos_idx = test_dataset.label_np[:, 0] == 1
        pos_candidates = test_dataset.candidate_np[pos_idx]
        all_candidates = np.unique(pos_candidates)
        num_all_candidates = len(all_candidates)
        # label will not be used when calculating metrics
        test_dataset.label_np = np.concatenate([np.zeros([len(pos_candidates), 2]) for _ in range(num_all_candidates)])
        ctx_np_pos = test_dataset.ctx_np[pos_idx]
        test_dataset.ctx_np = np.concatenate([ctx_np_pos for _ in range(num_all_candidates)])
        seq_np_pos = test_dataset.seq_np[pos_idx]
        test_dataset.seq_np = np.concatenate([seq_np_pos for _ in range(num_all_candidates)])
        vl_np_pos = test_dataset.vl_np[pos_idx]
        test_dataset.vl_np = np.concatenate([vl_np_pos for _ in range(num_all_candidates)])
        # loop thru every candidate
        test_dataset.candidate_np = np.concatenate([[all_candidates[i] for _ in range((len(pos_candidates)))]
                                                       for i in range(num_all_candidates)])

        if test_dataset.history_col:
            history_np_pos = test_dataset.history_np[pos_idx]
            test_dataset.history_np = np.concatenate([history_np_pos for _ in range(num_all_candidates)])
        test_dataset.len_data = len(test_dataset.ctx_np)
        self.pos_candidates = pos_candidates  # this will be used as target
        self.test_dataloader = DataLoader(test_dataset, batch_size=4096, num_workers=16, pin_memory=True)
        self.all_candidates = all_candidates
        self.num_all_candidates = num_all_candidates

    def on_validation_epoch_end(self, trainer, pl_module):
        # get predictions
        pl_module.eval()
        pl_module.to(self.device)
        preds = []
        with torch.no_grad():
            for (ctx, seq, vl, candidate), (_, _) in self.test_dataloader:
                out = pl_module(
                    ctx_in=[c.to(self.device) for c in ctx],
                    seq_in=seq.to(self.device),
                    vl_in=vl.to(self.device),
                    candidate_in=candidate.to(self.device),
                    seq_history=None
                )
                preds.append(out[0].cpu().numpy())
        preds = np.concatenate(preds)
        # count top k
        top_k_purchased_total = top_k_unpurchased_total = 0
        for input_idx in range(len(self.pos_candidates)):
            pred_top_k_idx = np.argsort(preds[input_idx::len(self.pos_candidates)].flatten())[::-1]
            pred_top_k_candidate = [self.all_candidates[i] for i in pred_top_k_idx[:self.top_k]]
            if self.pos_candidates[input_idx] in pred_top_k_candidate:
                top_k_purchased_total += 1
            else:
                top_k_unpurchased_total += 1

        top_k_acc = top_k_purchased_total / (top_k_purchased_total + top_k_unpurchased_total)
        log_message = f"epoch: {pl_module.current_epoch+1}, " \
                      f"top_k_purchased_total: {top_k_purchased_total}, " \
                      f"top_k_unpurchased_total: {top_k_unpurchased_total}, " \
                      f"top_k_acc: {top_k_acc}"
        print(log_message)
        trainer.logger.main_logger.info(log_message)
        wandb.log({'top_k_acc': top_k_acc})
        wandb.log({'top_k_purchased_total': top_k_purchased_total})
        wandb.log({'top_k_unpurchased_total': top_k_unpurchased_total})


class MTWandbArtifactCallback(pl.callbacks.Callback):
    def __init__(self, wandb_run, lookup_output_dir, torch_model_params, pl_model_params, model_dir):
        self.wandb_run = wandb_run
        self.model_dir = model_dir
        self.lookup_output_dir = lookup_output_dir
        self.torch_model_params = torch_model_params
        pl_model_params["task1_criterion"] = str(pl_model_params["task1_criterion"])
        pl_model_params["task2_criterion"] = str(pl_model_params["task2_criterion"])
        pl_model_params["optimizer"] = str(pl_model_params["optimizer"])
        pl_model_params["multi_task_txt_kwargs"]["ctx_nums"] = [
            int(i) for i in pl_model_params["multi_task_txt_kwargs"]["ctx_nums"]
        ]
        self.pl_model_params = pl_model_params

    def on_train_start(self, trainer, pl_module):
        # src zip
        shutil.make_archive(f"{self.model_dir}/src", "zip", "../../src")
        src_artifact = wandb.Artifact("src", type="src_zip")
        src_artifact.add_file(f"{self.model_dir}/src.zip")
        self.wandb_run.log_artifact(src_artifact, aliases=[self.wandb_run.name])
        # lookup files
        lookup_artifact = wandb.Artifact("lookup_csv", type="lookup_tables")
        lookup_artifact.add_dir(self.lookup_output_dir)
        self.wandb_run.log_artifact(lookup_artifact, aliases=[self.wandb_run.name])
        # torch model params
        torch_model_params_path = os.path.join(self.model_dir, "torch_model_params.json")
        with open(torch_model_params_path, "w") as f:
            json.dump(self.torch_model_params, f)
        torch_params_artifact = wandb.Artifact("torch_params", type="params")
        torch_params_artifact.add_file(torch_model_params_path)
        self.wandb_run.log_artifact(torch_params_artifact, aliases=[self.wandb_run.name])
        # pl model params
        pl_model_params_path = os.path.join(self.model_dir, "pl_model_params.json")
        with open(pl_model_params_path, "w") as f:
            json.dump(self.pl_model_params, f)
        pl_params_artifact = wandb.Artifact("pl_params", type="params")
        pl_params_artifact.add_file(pl_model_params_path)
        self.wandb_run.log_artifact(pl_params_artifact, aliases=[self.wandb_run.name])

    def on_train_epoch_end(self, trainer, pl_module):
        # torch model
        model_path = os.path.join(self.model_dir, f"model.pt")
        torch.save(pl_module.multi_task.state_dict(), model_path)
        torch_model_artifact = wandb.Artifact(f"torch_model_epoch_{pl_module.current_epoch}", type="model")
        torch_model_artifact.add_file(model_path)
        self.wandb_run.log_artifact(torch_model_artifact, aliases=[self.wandb_run.name])

    def on_train_end(self, trainer, pl_module):
        # pl model
        model_path = os.path.join(self.model_dir, f"model.pt")
        torch.save(pl_module.state_dict(), model_path)
        pl_model_artifact = wandb.Artifact(f"pl_model_epoch_{pl_module.current_epoch}", type="model")
        pl_model_artifact.add_file(model_path)
        self.wandb_run.log_artifact(pl_model_artifact, aliases=[self.wandb_run.name])


class MTTensorboardCallback(TensorboardCallbackBase):
    def __init__(self, tensorboard_logger):
        super().__init__(tensorboard_logger)

    def update_metrics_scalar(self, trainer, pl_module, phase):
        epoch = pl_module.current_epoch
        trainer_logger = trainer.logger
        trainer_loss_logger = trainer_logger.loss_logger
        trainer_task1_logger = trainer_logger.task1_logger
        trainer_task2_logger = trainer_logger.task2_logger
        self.tensorboard_logger.experiment.add_scalar(f"loss1/{phase}", trainer_loss_logger.get("loss1"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"loss2/{phase}", trainer_loss_logger.get("loss2"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"loss/{phase}", trainer_loss_logger.get("loss"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"acc/{phase}", trainer_task1_logger.get("acc"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"f1/{phase}", trainer_task1_logger.get("f1"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"mae/{phase}", trainer_task2_logger.get("mae"), epoch)
        self.tensorboard_logger.experiment.add_scalar(f"mse/{phase}", trainer_task2_logger.get("mse"), epoch)


class MTContextAttnVisCallback(AttnVisCallbackBase):
    def __init__(self, transformer_prefix, cols, sample_inputs, device):
        super().__init__(transformer_prefix, cols, sample_inputs, device)

    def get_attn_weights(self, pl_module):
        ctx_embedding_layers = pl_module.multi_task.shared_bottom.context_transformer.ctx_embedding
        attn_layer = pl_module.multi_task.shared_bottom.context_transformer.ctx_encoder.layers[0].self_attn
        ctx_embedding_list = [ctx_embedding_layers[i](input_ctx.to(self.device)).unsqueeze(1)
                              for i, input_ctx in enumerate(self.sample_inputs)]
        ctx_out = torch.cat(ctx_embedding_list, dim=1)
        _, attn_weight = attn_layer(ctx_out, ctx_out, ctx_out, need_weights=True)
        avg_attn_weight = attn_weight.mean(0).mean(0).detach().tolist()
        return avg_attn_weight


class MTSequenceAttnVisCallback(AttnVisCallbackBase):
    def __init__(self, transformer_prefix, cols, sample_inputs, device):
        super().__init__(transformer_prefix, cols, sample_inputs, device)

    def get_attn_weights(self, pl_module):
        seq_in = self.sample_inputs[0].to(self.device)
        vl_in = self.sample_inputs[1].to(self.device)
        sequence_transformer_layer = pl_module.multi_task.shared_bottom.sequence_transformer
        attn_layer = pl_module.multi_task.shared_bottom.sequence_transformer.seq_encoder.layers[0].self_attn
        seq_out = sequence_transformer_layer.seq_embedding(seq_in.long())
        seq_out = seq_out * math.sqrt(sequence_transformer_layer.seq_embed_size)
        seq_out = sequence_transformer_layer.pos_encoder(seq_out)
        mask = sequence_transformer_layer.create_key_padding_mask(seq_in=seq_in, valid_length=vl_in)
        _, attn_weight = attn_layer(seq_out, seq_out, seq_out, key_padding_mask=mask, need_weights=True)
        avg_attn_weight = attn_weight.mean(0).mean(0).detach().tolist()
        return avg_attn_weight


class OrderDataset(Dataset):
    """
    If from_s3 is True, will read data from s3. s3_bucket, s3_prefix_data and s3_prefix_lookups are required.
    Otherwise will read local data. local_data_data_dir and local_lookup_dirs are required.
    """
    def __init__(self, ctx_cols, seq_col, vl_col, candidate_col, history_col, label_cols,
                 from_s3=True, local_data_dir=None, local_lookup_dirs=None,
                 s3_bucket=None, s3_prefix_data=None, s3_prefix_lookups=None,
                 parts_num=9999, data_file_format="json", lookup_file_format="json"):
        self.ctx_cols = ctx_cols
        self.seq_col = seq_col
        self.vl_col = vl_col
        self.candidate_col = candidate_col
        self.history_col = history_col
        self.label_cols = label_cols
        self.local_data_dir = local_data_dir
        self.local_lookup_dirs = local_lookup_dirs
        self.s3_bucket = s3_bucket
        self.s3_prefix_data = s3_prefix_data
        self.s3_prefix_lookups = s3_prefix_lookups
        self.parts_num = parts_num
        self.data_file_format = data_file_format
        self.lookup_file_format = lookup_file_format
        # load data
        self.ctx_np = []
        self.seq_np = []
        self.vl_np = []
        self.candidate_np = []
        self.label_np = []
        self.history_np = []
        self.lookups = {}
        if from_s3:
            assert s3_bucket and s3_prefix_data and s3_prefix_lookups
            self._load_lookups_from_s3()
            self._load_data_from_s3()
        else:
            assert local_data_dir, local_lookup_dirs
            self._load_lookups_from_local()
            self._load_data_from_local()

        # convert dataframe to numpy
        self.ctx_np = np.concatenate(self.ctx_np, axis=0)
        self.seq_np = np.concatenate(self.seq_np, axis=0)
        self.vl_np = np.concatenate(self.vl_np, axis=0)
        self.candidate_np = np.concatenate(self.candidate_np, axis=0)
        self.label_np = np.concatenate(self.label_np, axis=0)
        if self.history_col:
            self.history_np = np.concatenate(self.history_np, axis=0)
        self.len_data = len(self.ctx_np)

    @staticmethod
    def _read_df(path, file_format):
        if file_format == "json":
            return pd.read_json(path, lines=True)
        elif file_format == "csv":
            return pd.read_csv(path)

    def _append_np_from_df(self, df):
        self.ctx_np.append(np.array(df[self.ctx_cols]))
        self.seq_np.append(np.array(df[self.seq_col].to_list()))
        self.vl_np.append(np.array(df[self.vl_col]))
        self.candidate_np.append(np.array(df[self.candidate_col]))
        self.label_np.append(np.array(df[self.label_cols].astype(float)))
        if self.history_col:
            self.history_np.append(np.array(df[self.history_col]))

    def _load_data_from_s3(self):
        keys = []
        s3_client = boto3.client('s3')
        resp = s3_client.list_objects_v2(Bucket=self.s3_bucket, Prefix=self.s3_prefix_data)
        for obj in resp['Contents']:
            if obj['Key'].endswith(self.data_file_format):
                keys.append(obj['Key'])
        for i, key in enumerate(keys):
            obj = s3_client.get_object(Bucket=self.s3_bucket, Key=key)
            df = self._read_df(obj['Body'], file_format=self.data_file_format)
            self._append_np_from_df(df)
            if i > self.parts_num:
                break

    def _load_data_from_local(self):
        f_paths = [
            os.path.join(self.local_data_dir, f_name)
            for f_name in os.listdir(self.local_data_dir)
            if f_name.endswith(self.data_file_format)
        ]
        for i, f_path in enumerate(f_paths):
            df = self._read_df(f_path, self.data_file_format)
            self._append_np_from_df(df)
            if i > self.parts_num:
                break

    def _load_lookups_from_s3(self):
        for s3_prefix in self.s3_prefix_lookups:
            lookup_name = os.path.split(s3_prefix[:-1])[1] if s3_prefix.endswith("/") else os.path.split(s3_prefix)[1]
            keys = []
            s3_client = boto3.client('s3')
            resp = s3_client.list_objects_v2(Bucket=self.s3_bucket, Prefix=s3_prefix)
            for obj in resp['Contents']:
                if obj['Key'].endswith(self.lookup_file_format):
                    keys.append(obj['Key'])
            assert len(keys) == 1, "should only have one lookup table per folder"
            obj = s3_client.get_object(Bucket=self.s3_bucket, Key=keys[0])
            df = self._read_df(obj['Body'], file_format=self.lookup_file_format)
            self.lookups[lookup_name] = df

    def _load_lookups_from_local(self):
        for local_dir in self.local_lookup_dirs:
            lookup_name = os.path.split(local_dir[:-1])[1] if local_dir.endswith("/") else os.path.split(local_dir)[1]
            local_paths = [
                file_name for file_name in os.listdir(local_dir)
                if file_name.endswith(self.lookup_file_format)
            ]
            assert len(local_paths) == 1, "should only have one lookup table per folder"
            df = self._read_df(os.path.join(local_dir, local_paths[0]), file_format=self.lookup_file_format)
            self.lookups[lookup_name] = df

    def __len__(self):
        return self.len_data

    def __getitem__(self, idx):
        ctx = [i for i in self.ctx_np[idx]]
        seq = torch.tensor(self.seq_np[idx])
        vl = self.vl_np[idx]
        candidate = self.candidate_np[idx]
        labels = self.label_np[idx]
        if self.history_col:
            seq_history = torch.tensor(self.seq_np[idx])
            data = (ctx, seq, vl, candidate, seq_history)
        else:
            data = (ctx, seq, vl, candidate)
        return data, (labels[0], labels[1])

    def get_feature_num(self, col):
        # TODO: implement via lookup dict
        # data_col = self.data[col]
        # if data_col.dtype == "int64":
        #     return data_col.astype(int).max() + 1
        if col == self.ctx_cols[0]:
            return self.ctx_np[:, 0].astype(int).max() + 1
        elif col == self.ctx_cols[1]:
            return self.ctx_np[:, 1].astype(int).max() + 1
        elif col == self.ctx_cols[2]:
            return self.ctx_np[:, 2].astype(int).max() + 1
        elif col == self.seq_col:
            return max(int(self.seq_np.reshape(-1).max()), int(self.candidate_np.reshape(-1).max())) + 1
        # elif data_col.dtype == "object":
        #     return int(np.array(data_col.to_list()).max()) + 1
        else:
            raise NotImplementedError()

    def export_lookups(self, local_dir, file_format):
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)
        for lookup_name, lookup_df in self.lookups.items():
            output_path = os.path.join(local_dir, f"{lookup_name}.{file_format}")
            if file_format == "json":
                lookup_df.to_json(output_path)
            elif file_format == "csv":
                lookup_df.to_csv(output_path, index=False)
            else:
                raise NotImplementedError()


class PLMultiTaskTxT(LightningModule):
    def __init__(self, task1_criterion, task2_criterion, optimizer, learning_rate,
                 loss_weights=(100, 1), log_interval=100,
                 multi_task_txt_model="MultiTaskTxT", multi_task_txt_kwargs=None):
        super().__init__()
        self.task1_criterion = task1_criterion
        self.task2_criterion = task2_criterion
        self.loss_weights = loss_weights
        self.log_interval = log_interval
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        if multi_task_txt_model == "MultiTaskTxT":
            self.multi_task = MultiTaskTxT(**multi_task_txt_kwargs)
        elif multi_task_txt_model == "MultiTaskTxE":
            self.multi_task = MultiTaskTxE(**multi_task_txt_kwargs)
        else:
            raise NotImplementedError()
        self.epoch_time = time.time()

    def forward(self, ctx_in, seq_in, vl_in, candidate_in, seq_history=None):
        return self.multi_task(ctx_in, seq_in, vl_in, candidate_in, seq_history)

    def forward_step(self, batch):
        (ctx, seq, vl, candidate), (label_pos, label_price) = batch
        out = self(
            ctx_in=ctx,
            seq_in=seq,
            vl_in=vl,
            candidate_in=candidate,
            seq_history=None
        )
        loss1 = self.task1_criterion(out[0].squeeze(), label_pos.float())
        loss2 = self.task2_criterion(out[1].squeeze(), label_price.float())
        loss = loss1 * self.loss_weights[0] + loss2 * self.loss_weights[1]
        y1_true = label_pos.float()
        y1_pred = (out[0].squeeze() > 0.5).float()
        y2_true = label_price.float()
        y2_pred = out[1].squeeze().float()
        return {"loss1": loss1.detach(), "loss2": loss2.detach(), "loss": loss,
                "y1_true": y1_true.detach(), "y1_pred": y1_pred.detach(),
                "y2_true": y2_true.detach(), "y2_pred": y2_pred.detach()}

    def on_train_epoch_start(self):
        self.reset_logger()

    def training_step(self, batch, batch_idx):
        return self.forward_step(batch)

    def training_step_end(self, step_output):
        self.update_logger(step_output)

    def on_validation_epoch_start(self):
        if not self.trainer.sanity_checking:
            self.write_logs("Train")
        self.reset_logger()

    def validation_step(self, batch, batch_idx):
        return self.forward_step(batch)

    def validation_step_end(self, step_output):
        self.update_logger(step_output)

    def on_validation_epoch_end(self):
        self.write_logs("Validation")

    def configure_optimizers(self):
        return self.optimizer(self.parameters(), lr=self.learning_rate)

    def update_logger(self, step_output):
        self.logger.task1_logger.update(metric_name=None, y_true=step_output["y1_true"], y_pred=step_output["y1_pred"])
        self.logger.task2_logger.update(metric_name=None, y_true=step_output["y2_true"], y_pred=step_output["y2_pred"])
        self.logger.loss_logger.update(metric_name="loss", y_true=step_output["loss"], y_pred=None)
        self.logger.loss_logger.update(metric_name="loss1", y_true=step_output["loss1"], y_pred=None)
        self.logger.loss_logger.update(metric_name="loss2", y_true=step_output["loss2"], y_pred=None)

    def reset_logger(self):
        self.epoch_time = time.time()
        self.logger.task1_logger.reset(metric_name=None)
        self.logger.task2_logger.reset(metric_name=None)
        self.logger.loss_logger.reset(metric_name=None)

    def write_logs(self, phase):
        log_str = "{} metrics Epoch {} Time elapsed {:.2f} - "\
                  "loss: {:.3f}, loss1: {:.3f}, loss2: {:.3f}, "\
                  "ACC: {:.2f}, F1: {:.2f}, MAE: {:.2f}, MSE: {:.2f}".format(
            phase, self.current_epoch + 1, time.time() - self.epoch_time,
            self.logger.loss_logger.get("loss"), self.logger.loss_logger.get("loss1"), self.logger.loss_logger.get("loss2"),
            self.logger.task1_logger.get("acc"), self.logger.task1_logger.get("f1"),
            self.logger.task2_logger.get("mae"), self.logger.task2_logger.get("mse"),
        )
        self.logger.main_logger.info(log_str)





def get_anitta_meal(purchase_history_id):
    # Anitta meal
    if "a04ac236-7876-40e6-816f-5451052e4448" in purchase_history_id:
        adhoc_offer = "XbQHVtlrnQ8pZHzm0qwzCp"  # $6 Impossible™ Whopper Combo Meal
    else:
        adhoc_offer = "n/a"
    return adhoc_offer


def get_ff_og_hb(purchase_history_id):
    # $1 Any Size French Fries, Onion Rings or Hash Browns purchaser
    ff = {"item_539", "item_540", "item_541"}  # French Fries
    og = {"item_686", "item_687", "item_688"}  # Onion Rings
    # hb = {"item_568", "item_569", "item_570"}  # Hash Browns
    if "23354564-7c6f-494f-b56c-8f156e530aaf" in purchase_history_id:
        purchase_cart = purchase_history_id["23354564-7c6f-494f-b56c-8f156e530aaf"]
        if len(og.intersection(purchase_cart)) > 0:
            adhoc_offer = "b2b74622-5746-4945-be1b-c0044757adf6"  # $1 Large Onion Rings
        elif len(ff.intersection(purchase_cart)) > 0:
            adhoc_offer = "99477cec-ca61-4851-98c2-4f87539e0a11"  # $1 Large French Fries
        else:
            adhoc_offer = "99477cec-ca61-4851-98c2-4f87539e0a11"  # $1 Large French Fries
    else:
        adhoc_offer = "99477cec-ca61-4851-98c2-4f87539e0a11"  # $1 Large French Fries
    return adhoc_offer


def get_sd_ic_hc(purchase_history_id):
    # $1 Any Size Hot Coffee, Iced Coffee, or Soft Drink
    sd = {
        "item_113", "item_376", "item_377", "item_379",
        # SM
        "item_43086", "item_52523", "item_43091", "item_52533", "item_43116", "item_43096", "item_52538", "item_43111",
        "item_43121", "item_43126", "item_43101", "item_52548", "item_43141", "item_43146", "item_43106", "item_52543",
        "item_43151", "item_43156",
        # MED
        "item_43087", "item_52524", "item_43092", "item_52534", "item_43097", "item_52539", "item_43107", "item_52544",
        "item_43112", "item_43117", "item_43122", "item_43102", "item_52549", "item_43122", "item_43127", "item_43142",
        "item_43147", "item_43152", "item_43157", "item_43137", "item_884", "item_879",
        # LG
        "item_43088", "item_52525", "item_43093", "item_52535", "item_43098", "item_52540", "item_43108", "item_52545",
        "item_43113", "item_43118", "item_43123", "item_43128", "item_43103", "item_52550", "item_43143", "item_43148",
        "item_43153", "item_43158",
    }
    ic = {
        # SM
        "item_580", "item_577", "item_583", "item_576", "item_42099",
        # MED
        "item_581", "item_578", "item_584", "item_575", "item_42112",
        # LG
        "item_582", "item_579", "item_585", "item_574", "item_42113",
    }
    hc = {
        # SM
        "item_49485", "item_49510",
        # MED
        "item_49508", "item_49511",
        # LG
        "item_49509", "item_49512",
    }
    if "e2e6f181-b069-4871-baab-aef10b90e394" in purchase_history_id:
        purchase_cart = purchase_history_id["e2e6f181-b069-4871-baab-aef10b90e394"]
        if len(sd.intersection(purchase_cart)) > 0:
            adhoc_offer = "29c77c38-0059-4981-9334-6936d03416ee"  # $1 Any Size Soft Drink
        elif len(ic.intersection(purchase_cart)) > 0:
            adhoc_offer = "ca0abc40-7f78-40c6-99cf-fe8482c6e300"  # $1 Small Iced Coffee
        elif len(hc.intersection(purchase_cart)) > 0:
            adhoc_offer = "1145dcc9-2fdc-4ee7-89a4-4f37cabc2374"  # $1 Any Size Coffee - Delivery Eligible
        else:
            adhoc_offer = "b2b74622-5746-4945-be1b-c0044757adf6"  # $1 Large Onion Rings
    else:
        adhoc_offer = "b2b74622-5746-4945-be1b-c0044757adf6"  # $1 Large Onion Rings
    return adhoc_offer


def get_adhoc_offers(purchase_history):
    purchase_history_id = {
        purchase.get("couponID", "n/a"): purchase.get("cart", [])
        for purchase in purchase_history
    }
    adhoc_offer_func = [
        # get_anitta_meal,
        get_ff_og_hb,
        get_sd_ic_hc,
    ]

    adhoc_offers = [
        func(purchase_history_id)
        for func in adhoc_offer_func
    ]
    adhoc_offers = [
        oid
        for oid in adhoc_offers
        if oid != "n/a"
    ]
    return adhoc_offers



import os
import logging
import warnings
import numpy as np
import pandas as pd
import mxnet as mx
from mxnet import gluon
from bert_serving.client import BertClient
import boto3


logger = logging.getLogger("offer_encoder")


# Helper functions

def load_s3_file(s3_bucket, s3_prefix, file_format="csv"):
    keys = []
    dfs = []
    s3 = boto3.client('s3')
    list_obj = s3.list_objects_v2(Bucket=s3_bucket, Prefix=s3_prefix)
    assert list_obj["KeyCount"] < 1000
    for content in list_obj["Contents"]:
        if content["Key"].endswith(file_format):
            keys.append(content["Key"])
    assert len(keys) == 1
    logger.info("Load %d %s files from %s" % (len(keys), file_format, os.path.dirname(keys[0])))
    for i, key in enumerate(keys):
        obj = s3.get_object(Bucket=s3_bucket, Key=key)
        if file_format == "csv":
            dfs.append(pd.read_csv(obj["Body"]))
        elif file_format == "json":
            dfs.append(pd.read_json(obj["Body"], lines=True))
        else:
            raise NotImplementedError()
    return pd.concat(dfs)


# Should be changed along with app_mxnet_inference.load_offer_id_from_pool
def load_offer_id_from_pool(offer_pool_path, active_offer_pool=None):
    offer_pool = pd.read_csv(offer_pool_path, dtype="str").dropna(subset=["Sanity ID"])
    offer_pool = offer_pool[~offer_pool["Sanity ID"].isin(["TBR", "TBD", "0", "5"])]
    offer_pool = offer_pool[~offer_pool["Sanity ID"].str.lower().str.contains("requested")]
    logger.info("Offer Pool without TBD and NA: %d" % len(offer_pool))

    # offer pools
    weekly_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &  # column removed
        # (offer_pool["National Offer Overlap?"] == "No") &  # column removed
        (offer_pool["Offer Type"].isin(["1_Personalized", "2_Personalized", "Lapsed_60", "Lapsed_90", "Default"]))
        # (offer_pool["L3"].isin(["Lapsed", "Segmented"])) &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    weekly_default_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &
        # (offer_pool["National Offer Overlap?"] == "No") &
        (offer_pool["Offer Type"] == "Default")
        # (offer_pool["L3"].isin(["Lapsed", "Segmented"])) &
        # (offer_pool["L4"].isin(["Default"]))
    ]["Sanity ID"].tolist()
    daily_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "MO&P") &
        # (offer_pool["National Offer Overlap?"] == "No") &
        (offer_pool["Offer Type"] == "Daily")
        # (offer_pool["L3"] == "Daily Deals") &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    delivery_offer_pool = offer_pool[
        # (offer_pool["Active/Inactive"] == "Active") &
        (offer_pool["Status"].isin(["Active", "Y"])) &
        # (offer_pool["Delivery/MO&P"] == "Delivery") &
        (offer_pool["Offer Type"] == "Delivery")
        # (offer_pool["L3"].isin(["Lapsed", "Segmented", "Daily Deals"])) &
        # (offer_pool["L4"].isin(["Default", "Personalized"]))
        ]["Sanity ID"].tolist()
    if active_offer_pool:
        weekly_offer_pool = [offer_id for offer_id in weekly_offer_pool if offer_id in active_offer_pool]
        weekly_default_pool = [offer_id for offer_id in weekly_default_pool if offer_id in active_offer_pool]
        daily_offer_pool = [offer_id for offer_id in daily_offer_pool if offer_id in active_offer_pool]
        delivery_offer_pool = [offer_id for offer_id in delivery_offer_pool if offer_id in active_offer_pool]
    inactive_pool = offer_pool[
        offer_pool["Sanity ID"].isin([i for i in offer_pool["Sanity ID"].values
                                      if i not in weekly_offer_pool + daily_offer_pool + delivery_offer_pool])
    ]
    logger.info("active weekly MO&P: %d" % len(weekly_offer_pool))
    logger.info("active weekly MO&P default: %d" % len(weekly_default_pool))
    logger.info("active daily MO&P: %d" % len(daily_offer_pool))
    logger.info("active weekly Delivery: %d" % len(delivery_offer_pool))
    offer_pools = {
        "weekly": weekly_offer_pool,
        "default": weekly_default_pool,
        "daily": daily_offer_pool,
        "delivery": delivery_offer_pool,
        "inactive": inactive_pool,
    }
    return offer_pools


def upload_file(file_name, bucket, object_name=None):
    """Upload a file to an S3 bucket

    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_name: S3 object name. If not specified then file_name is used
    :return: True if file was uploaded, else False
    """

    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    s3_client = boto3.client('s3')
    try:
        response = s3_client.upload_file(file_name, bucket, object_name)
    except Exception as e:
        logging.error(e)
        return False
    return True


# Main functions

def load_offer_all(s3_bucket, s3_prefix_offer_mapping, s3_prefix_offer_history, offer_pool_path):
    """
    offer_pool_all:
    offer_id = Unique(offer_history + weekly_offer + daily_offer + delivery_offer)
    offer_name = offer_id --offer_id_mapping-> offer_name
    """
    offer_id_mapping = load_s3_file(s3_bucket=s3_bucket, s3_prefix=s3_prefix_offer_mapping)
    offer_id_mapping = offer_id_mapping.rename(columns={"id": "offer_id",
                                                        "name": "offer_name",
                                                        "localizedImageUrl": "offer_img"})
    # TODO: supplementary offer mapping
    offer_id_mapping_supp = pd.read_csv("./data/archived/sanity_prod_2021-10-12.csv").rename(
        columns={"id": "offer_id", "name": "offer_name", "localizedImageUrl": "offer_img"})
    offer_id_mapping = pd.concat([
        offer_id_mapping[["offer_id", "offer_name", "offer_img"]],
        offer_id_mapping_supp[["offer_id", "offer_name", "offer_img"]]
    ]).drop_duplicates(subset=["offer_id"], keep="first")

    offer_history = load_s3_file(
        s3_bucket=s3_bucket,
        s3_prefix=s3_prefix_offer_history,
        file_format="json").drop(columns=["offer_name"])
    logger.info("Load %d history offers from train/eval data" % len(offer_history))

    offer_pools = load_offer_id_from_pool(offer_pool_path=offer_pool_path,
                                          active_offer_pool=offer_id_mapping["offer_id"].tolist())
    active_pool = offer_pools["weekly"] + offer_pools["daily"] + offer_pools["delivery"]
    active_pool_not_in_history = [offer_id for offer_id in active_pool
                                  if offer_id not in offer_history["offer_id"].values]
    history_max_idx = offer_history["offer_idx"].max()
    offer_pool_not_in_history = pd.DataFrame(
        data=zip(active_pool_not_in_history,
                 range(history_max_idx + 1, history_max_idx + 1 + len(active_pool_not_in_history)),
                 ["false"] * len(active_pool_not_in_history)),
        columns=["offer_id", "offer_idx", "in_training"]
    )
    logger.info("Load %d (%d) active offers from offer pool (exclusively)" % (len(active_pool),
                                                                              len(active_pool_not_in_history)))

    offer_pool_all = pd.concat([offer_history, offer_pool_not_in_history]).merge(
        offer_id_mapping[["offer_id", "offer_name", "offer_img"]],
        on="offer_id",
        how="inner").sort_values(by=["offer_idx"]).reset_index(drop=True)
    logger.info("Process to %d distinct offers in offer features lookup" % len(offer_pool_all))
    assert len(offer_history) + len(offer_pool_not_in_history) == len(offer_pool_all), "Rows dropped during merging"
    assert offer_pool_all["offer_id"].isna().sum() == 0, "NA found in offer_id"
    assert offer_pool_all.dropna(subset=["offer_id"]).duplicated(subset=["offer_id"]).sum() == 0, "Duplicated offer_id"
    assert offer_pool_all["offer_name"].isna().sum() == 0, "NA found in offer_name"

    # add index 0
    offer_pool_all = pd.concat([pd.DataFrame(data=[["n/a", 0, "n/a", "n/a", "n/a"]],
                                             columns=offer_pool_all.columns), offer_pool_all])
    logger.info("Add 1 empty offer to pool")
    return offer_pool_all


def get_bert_features(offer_pool_all):
    logger.info("Found %d NA or empty offer names" % (offer_pool_all["offer_name"].isna().sum()
                                                      + sum(offer_pool_all["offer_name"] == "")))
    offer_names_lower = offer_pool_all[offer_pool_all["offer_name"] != "n/a"]["offer_name"].str.lower().tolist()
    bc = BertClient(timeout=10000)
    bert_features = bc.encode(offer_names_lower)
    bert_features = np.concatenate([np.zeros((1, bert_features.shape[1]), dtype="float32"),
                                    bert_features])
    logger.info("Generate %s BERT features" % (bert_features.shape,))
    return bert_features


def get_resnet_features(offer_pool_all, local_img_dir="./data/offer_img/"):
    if not os.path.exists(local_img_dir):
        os.makedirs(local_img_dir)
    ctx = mx.cpu()
    net = gluon.model_zoo.vision.resnet50_v1(pretrained=True, ctx=ctx)
    net.hybridize(static_alloc=True, static_shape=True)
    resnet_features = []
    warnings.filterwarnings("ignore")
    for i, (offer_id, offer_url) in offer_pool_all[["offer_id", "offer_img"]].iterrows():
        try:
            img_path = gluon.utils.download(url="https://" + offer_url,
                                            path=os.path.join(local_img_dir, "%s.png" % offer_id),
                                            overwrite=True)
        except Exception as e:
            img_path = ""
        try:
            img = mx.image.imread(img_path)
        except Exception as e:
            if img_path == "":
                logger.info("Offer image url NA: %s" % offer_id)
            else:
                logger.info("Offer image read failed: %s" % offer_id)
            img = mx.ndarray.zeros((224, 224, 3), dtype="uint8")
        img = mx.image.imresize(img, 224, 224)  # resize
        img = mx.image.color_normalize(img.astype(dtype='float32') / 255,
                                       mean=mx.nd.array([0.485, 0.456, 0.406]),
                                       std=mx.nd.array([0.229, 0.224, 0.225]))  # normalize
        img = img.transpose((2, 0, 1))  # channel first
        img = img.expand_dims(axis=0)  # batchify
        img = img.as_in_context(ctx)
        img_features = net.features(img).asnumpy().flatten()
        resnet_features.append(img_features)
    resnet_features = np.array(resnet_features)
    logger.info("Generate %s ResNet features" % (resnet_features.shape,))
    return resnet_features


def save_embedding(net, train_iter, ctx, offer_pool_all, local_net_path, local_offer_lookup_path,
                   local_embed_csv_path, local_embed_json_path, s3_bucket, s3_embed_csv_path, s3_embed_json_path):
    train_iter.reset()
    data_x = train_iter.next().data[0].as_in_context(ctx)
    data_h = net.encoder(data_x)
    data_mu = mx.nd.split(data_h, axis=1, num_outputs=2)[0]
    # data_lv = mx.nd.split(data_h, axis=1, num_outputs=2)[1]
    net.save_parameters(local_net_path)
    logger.info("Save VAE net to %s" % local_net_path)

    # embed csv
    offer_embed_embedding_csv = pd.DataFrame(data=data_mu.asnumpy(),
                                             columns=["embed_%d" % i for i in range(data_mu.shape[1])])
    offer_embed_csv = offer_pool_all.merge(offer_embed_embedding_csv, left_index=True, right_index=True, how="left")
    offer_embed_csv.to_csv(local_embed_csv_path, index=False)
    offer_embed_csv.to_csv(local_offer_lookup_path, index=False)
    upload_file(local_embed_csv_path, s3_bucket, s3_embed_csv_path)
    logger.info("Save offer embed to %s" % local_embed_csv_path)
    logger.info("Save offer lookup to %s" % local_offer_lookup_path)
    logger.info("Upload offer embed to s3: %s/%s" % (s3_bucket, s3_embed_csv_path))

    # embed json
    offer_embed_embedding_json = pd.DataFrame(data=[[embed.tolist()] for embed in data_mu.asnumpy()], columns=["embed"])
    offer_embed_json = offer_pool_all.merge(offer_embed_embedding_json, left_index=True, right_index=True, how="left")
    offer_embed_json.to_json(local_embed_json_path, orient="records", lines=True)
    upload_file(local_embed_json_path, s3_bucket, s3_embed_json_path)
    logger.info("Save offer embed to %s" % local_embed_json_path)
    logger.info("Upload offer embed to s3 :%s/%s" % (s3_bucket, s3_embed_json_path))
    return data_mu










import torch
import torch.nn as nn
from collections import OrderedDict


class MLP(nn.Module):
    def __init__(self, input_size, hidden_sizes, activation, last_layer_activation=False):
        super().__init__()
        self.num_features = (input_size,) + tuple(hidden_sizes)
        self.net = nn.Sequential()
        if isinstance(activation, str):
            act_fn = _get_activation_fn(activation, dim=1)
        else:
            act_fn = activation
        n = 0
        for i, (curr_in, curr_out) in enumerate(zip(self.num_features[0:-1], self.num_features[1:])):
            self.net.add_module(str(n), nn.Linear(in_features=curr_in, out_features=curr_out))
            n += 1
            if last_layer_activation or i < (len(hidden_sizes) - 1):
                self.net.add_module(str(n), act_fn)
                n += 1

    def forward(self, x):
        x = self.net(x)
        return x


class MMoE(nn.Module):
    def __init__(self, input_size, expert_num, expert_hidden_sizes,
                 task_num, task_hidden_sizes, task_last_activations):
        super().__init__()
        assert task_num == len(task_hidden_sizes) == len(task_last_activations), "task num does not match"

        self.input_size = input_size
        self.expert_num = expert_num
        self.expert_hidden_sizes = tuple(expert_hidden_sizes)
        self.task_num = task_num
        self.task_hidden_sizes = tuple(task_hidden_sizes)
        self.task_last_activations = nn.ModuleList([
            _get_activation_fn(act) if act is None or isinstance(act, str) else act
            for act in task_last_activations
        ])
        # self.tower_size = tower_size
        # self._param_expert = []

        self.experts = nn.ModuleDict()
        for i in range(self.expert_num):
            self.experts[f"expert_mlp_{i}"] = MLP(
                input_size=input_size,
                hidden_sizes=self.expert_hidden_sizes,
                activation="relu",
                last_layer_activation=True,
            )

        self.gates = nn.ModuleDict()
        self.tasks = nn.ModuleDict()
        for i in range(self.task_num):
            self.gates[f"gate_mlp_{i}"] = MLP(
                input_size=input_size,
                hidden_sizes=[expert_num],
                activation="softmax",
                last_layer_activation=True,
            )  # TODO: UserWarning: Implicit dimension choice for softmax has been deprecated. Change the call to include dim=X as an argument.
            self.tasks[f"task_mlp_{i}"] = MLP(
                input_size=self.expert_hidden_sizes[-1],
                hidden_sizes=self.task_hidden_sizes[i],
                activation="relu",
                last_layer_activation=False
            )

    def forward(self, features_in):
        expert_outs_list = []
        for i in range(0, self.expert_num):
            expert_out = self.experts[f"expert_mlp_{i}"](features_in)
            expert_outs_list.append(expert_out)  # (B, hidden] * E
        expert_outs = torch.stack(expert_outs_list, dim=1)  # [B, E, expert_hidden]

        task_outs_list = []
        for i in range(self.task_num):
            gate_weight = self.gates[f"gate_mlp_{i}"](features_in)  # [B, E]
            gate_weight = gate_weight.unsqueeze(2)  # [B, E, 1]
            gate_out = expert_outs.mul(gate_weight).sum(dim=1)  # [B, expert_hidden]
            task_out = self.tasks[f"task_mlp_{i}"](gate_out)
            task_out = self.task_last_activations[i](task_out)  # [B, task_hidden]
            # task_out = task_out.clamp(min=1e-15, max=1.0 - 1e-15  # TODO: need clip?
            task_outs_list.append(task_out)
        return task_outs_list


def _get_activation_fn(activation, dim=None):
    if activation is None or activation.lower() == "none":
        return nn.Identity()
    elif activation == "relu":
        return nn.ReLU()
    elif activation == "gelu":
        return nn.GELU()
    elif activation == "sigmoid":
        return nn.Sigmoid()
    elif activation == "softmax":
        return nn.Softmax(dim=dim)
    elif activation == "tanh":
        return nn.Tanh()
    # elif activation == "leakyrelu":
    #     return nn.LeakyReLU()
    # elif activation == "softrelu":
    #     return nn.Softplus()

    raise RuntimeError("activation should be relu/gelu/sigmoid/softmax/tanh, not {}".format(activation))



import os
import random
import json
import pandas as pd
import torch
import boto3
from botocore.config import Config
from src.torch_layers import TxTFeaturesExtractor


def generate_lookup_dict(df, key_col, value_col):
    """Generate int2int lookup dict from given dataframe."""
    return df[[key_col, value_col]].set_index(key_col).to_dict()[value_col]


def load_lookup_model(model_name, offer_lookup_dynamodb_table="digitalOfferLookup"):
    """
    Load lookup tables from local directory
    :param model_name: root directory all model files
    :param offer_lookup_dynamodb_table: table name of offer_lookup table in DynamoDB
    :return: lookup tables and TxT model
    """
    # read offer_lookup from DynamoDB
    my_config = Config(region_name='us-east-1', connect_timeout=5, read_timeout=5, retries={'max_attempts': 10})
    dynamodb = boto3.resource('dynamodb', config=my_config)
    digital_offer_lookup = dynamodb.Table(offer_lookup_dynamodb_table)
    response = digital_offer_lookup.scan()
    items = response['Items']
    while 'LastEvaluatedKey' in response:
        response = digital_offer_lookup.scan(ExclusiveStartKey=response['LastEvaluatedKey'])
        items.extend(response['Items'])
    offer_lookup = pd.DataFrame({'data': items})
    offer_lookup.loc[:, "OFFER_ID"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("OFFER_ID"))
    offer_lookup.loc[:, "OFFER_IDX"] = offer_lookup.loc[:, "data"].apply(lambda x: int(x.get("OFFER_IDX")))
    offer_lookup.loc[:, "SENDABLE"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("SENDABLE"))
    offer_lookup.loc[:, "BREAKFAST"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("BREAKFAST"))
    offer_lookup.loc[:, "LAPSED_90"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("LAPSED_90"))
    offer_lookup.loc[:, "OFFER_STRUCTURE"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("OFFER_STRUCTURE"))
    offer_lookup.loc[:, "OFFER_URL"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("OFFER_URL"))
    offer_lookup.loc[:, "GPM_OFFER"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("GPM_OFFER"))
    offer_lookup.loc[:, "OFFER_NAME_FINALIZED"] = offer_lookup.loc[:, "data"].apply(
        lambda x: x.get("OFFER_NAME_FINALIZED")
    )
    offer_lookup.loc[:, "OFFER_NAME_FORMATTED"] = offer_lookup.loc[:, "data"].apply(
        lambda x: x.get("OFFER_NAME_FORMATTED")
    )
    offer_lookup.loc[:, "OFFER_NAME"] = offer_lookup.loc[:, "data"].apply(lambda x: x.get("OFFER_NAME"))
    offer_lookup = offer_lookup.drop(columns=["data"])
    offer_lookup_img = offer_lookup[["OFFER_ID", "OFFER_NAME_FORMATTED", "OFFER_URL"]].copy()
    # print("Load offer lookup table from DynamoDB table %s" % offer_lookup_dynamodb_table)

    # Paths
    model_path = os.path.join("models/", model_name, "model.pt")
    platform_lookup_path = os.path.join("models/", model_name, "platform_lookup.csv")
    dma_lookup_path = os.path.join("models/", model_name, "dma_lookup.csv")
    device_carrier_lookup_path = os.path.join("models/", model_name, "device_carrier_lookup.csv")
    device_model_lookup_path = os.path.join("models/", model_name, "device_model_lookup.csv")

    # Load lookup table
    platform_lookup = pd.read_csv(platform_lookup_path)
    dma_lookup = pd.read_csv(dma_lookup_path)
    device_carrier_lookup = pd.read_csv(device_carrier_lookup_path)
    device_model_lookup = pd.read_csv(device_model_lookup_path)
    idx2offerid = generate_lookup_dict(offer_lookup, key_col="OFFER_IDX", value_col="OFFER_ID")
    idx2offername = generate_lookup_dict(offer_lookup, key_col="OFFER_IDX", value_col="OFFER_NAME_FINALIZED")
    offerid2idx = generate_lookup_dict(offer_lookup, key_col="OFFER_ID", value_col="OFFER_IDX")
    offername2idx = generate_lookup_dict(offer_lookup, key_col="OFFER_NAME_FINALIZED", value_col="OFFER_IDX")
    platform2idx = generate_lookup_dict(platform_lookup, key_col="PLATFORM", value_col="PLATFORM_IDX")
    dma2idx = generate_lookup_dict(dma_lookup, key_col="DMA", value_col="DMA_IDX")
    carrier2idx = generate_lookup_dict(device_carrier_lookup, key_col="DEVICE_CARRIER", value_col="DEVICE_CARRIER_IDX")
    model2idx = generate_lookup_dict(device_model_lookup, key_col="DEVICE_MODEL", value_col="DEVICE_MODEL_IDX")
    # print("Load platform lookup table from %s" % platform_lookup_path)
    # print("Load dma lookup table from %s" % dma_lookup_path)
    # print("Load device_carrier lookup table from %s" % device_carrier_lookup_path)
    # print("Load device_model lookup table from %s" % device_model_lookup_path)
    lookup_tables = {
        "idx2offerid": idx2offerid,
        "idx2offername": idx2offername,
        "offerid2idx": offerid2idx,
        "offername2idx": offername2idx,
        "platform2idx": platform2idx,
        "dma2idx": dma2idx,
        "carrier2idx": carrier2idx,
        "model2idx": model2idx,
    }

    # Load model
    model = torch.load(model_path, map_location=torch.device("cpu"))
    model.eval()
    # print("Load model from %s" % model_path)
    return offer_lookup_img, offer_lookup.set_index("OFFER_ID"), lookup_tables, model


def transform_raw_to_history(raw_history):
    """
    Transform user raw_history to processed recent history
    :param raw_history: list of user offer redeemed events
    :return: processed recent history of the user
    """
    # filtered_history = [v for v in raw_history if v.get("orderType") == "restaurant order"]
    sorted_history = sorted(raw_history, key=lambda v: v["timeStamp"])
    sorted_offer_id = list(map(lambda v: v["couponID"], sorted_history))
    len_history = len(sorted_offer_id)
    processed_history = {
        "platform": sorted_history[-1].get("platform", "n/a"),
        # "dma": "Miami-Ft. Lauderdale, FL",
        "device_carrier": sorted_history[-1].get("carrier", "n/a"),
        "device_model": sorted_history[-1].get("deviceModel", "n/a"),
        "offer_id": sorted_offer_id[max(0, len_history - 5):] + ["n/a"] * max(0, 5 - len_history),
    }
    return processed_history


def transform_history_to_model_in(history, offerid2idx, platform2idx, dma2idx, carrier2idx, model2idx):
    """
    Transform readable recent history to indexed model_input
    :param history: processed recent history of the user
    :param offerid2idx: dict of offer_id => idx
    :param platform2idx: dict of platform => idx
    :param dma2idx:
    :param carrier2idx:
    :param model2idx:
    :return:
    """
    # TODO: all in-coming str are lowercased
    offer_idx = [offerid2idx.get(offer_id, 0) for offer_id in history["offer_id"]]
    platform_idx = platform2idx.get(history["platform"], 1)  # web is 1
    # dma_idx = dma2idx.get(history["dma"], 0)
    device_carrier_idx = carrier2idx.get(history["device_carrier"], 0)
    device_model_idx = model2idx.get(history["device_model"], 0)

    context_in = [torch.tensor([[context]])
                  for context in [platform_idx, device_carrier_idx, device_model_idx]]  # TODO: no DMA for now
    sequence_in = torch.tensor([offer_idx])
    return context_in, sequence_in


def transform_model_out_to_offer_id(model_out, idx2offerid):
    sorted_idx = model_out.argsort(descending=True).flatten()
    sorted_offer_id = [idx2offerid.get(idx.item(), 0) for idx in sorted_idx]
    return [offer_id for offer_id in sorted_offer_id if offer_id != 0]


def is_lapsed(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["LAPSED_90"] == "Y"


def is_structure_bundle(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["OFFER_STRUCTURE"] == "Bundle"


def is_structure_sides(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["OFFER_STRUCTURE"] == "Sides / Desserts"


def is_breakfast(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["BREAKFAST"] == "Y"


def is_discount_low(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["GPM_OFFER"] == "L"


def is_discount_medium(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["GPM_OFFER"] == "M"


def is_discount_high(offer_lookup, offer_id):
    return offer_lookup.loc[offer_id]["GPM_OFFER"] == "H"


def get_offer_attributes(offer_lookup, offer_group_lookup, offer_id):
    offer_attributes = {
        "offer_group": offer_group_lookup[offer_id],
        "breakfast": None,
        "discount_level": None,
        "offer_structure": None,
    }
    # breakfast
    if is_breakfast(offer_lookup, offer_id):
        offer_attributes["breakfast"] = "br"
    else:
        offer_attributes["breakfast"] = "nbr"
    # discount_level
    if is_discount_high(offer_lookup, offer_id):
        offer_attributes["discount_level"] = "h"
    elif is_discount_medium(offer_lookup, offer_id):
        offer_attributes["discount_level"] = "m"
    else:
        offer_attributes["discount_level"] = "l"
    # offer_structure
    if is_structure_bundle(offer_lookup, offer_id):
        offer_attributes["offer_structure"] = "bundle"
    elif is_structure_sides(offer_lookup, offer_id):
        offer_attributes["offer_structure"] = "sides"
    else:
        offer_attributes["offer_structure"] = "other"
    return offer_attributes


def update_cluster_rules(topk_offer_attr, cluster_rules, offer_lookup,
                         offer_group_lookup, offer_id, use_cluster_rules=True):
    offer_attributes = get_offer_attributes(offer_lookup, offer_group_lookup, offer_id)
    group = offer_attributes["offer_group"]
    br_nbr = offer_attributes["breakfast"]
    h_m_l = offer_attributes["discount_level"]
    structure = offer_attributes["offer_structure"]
    topk_offer_attr = topk_offer_attr.copy()
    if use_cluster_rules:
        satisfy_rule = (topk_offer_attr[br_nbr] < cluster_rules[br_nbr]
                        and topk_offer_attr[h_m_l] < cluster_rules[h_m_l]
                        and (group == 0 or group not in topk_offer_attr["groups"]))
    else:
        satisfy_rule = (group == 0 or group not in topk_offer_attr["groups"])
    if satisfy_rule:
        topk_offer_attr[br_nbr] += 1
        topk_offer_attr[h_m_l] += 1
        topk_offer_attr[structure] += 1
        topk_offer_attr["groups"].append(group)
    return satisfy_rule, topk_offer_attr


def transform_offer_attr(offer_attr):
    offer_attr_log = "br-{},nbr-{},h-{},m-{},l-{}|bundle-{},sides-{},other-{}|groups-{}".format(
        offer_attr["br"],
        offer_attr["nbr"],
        offer_attr["h"],
        offer_attr["m"],
        offer_attr["l"],
        offer_attr["bundle"],
        offer_attr["sides"],
        offer_attr["other"],
        str(offer_attr["groups"])
    )
    return offer_attr_log


def filter_raw_by_rules(sorted_offer_id, lapsed, offer_lookup, topk, lapsed_rule, cluster_rules, offer_group_lookup):
    """
    Filter sorted model_out by rules generated by user cluster
    :param offer_group_lookup:
    :param sorted_offer_id: list of sorted offer_id, descending
    :param lapsed: int 0, 30, 60 or 90
    :param offer_lookup: offer_lookup table containing offer info
    :param topk: number of offers returned
    :param lapsed_rule: number of lapsed offer assigned to lapsed user
    :param cluster_rules: [(br, nbr), (H, M, L)]
    :return: list of sorted and filtered offer_id that satisfies all given rules
    """
    topk_offer_id = []
    topk_offer_attr = {
        "br": 0,
        "nbr": 0,
        "h": 0,
        "m": 0,
        "l": 0,
        "bundle": 0,
        "sides": 0,
        "other": 0,
        "groups": [],
    }
    # filter SENDABLE
    offer_id_sendable = offer_lookup[offer_lookup["SENDABLE"] == "Y"].index.values
    # TODO: add active filter
    offer_id_sendable_sorted = list(filter(lambda x: x in offer_id_sendable, sorted_offer_id))

    # filter LAPSED
    if lapsed == "90":
        lapsed_count = 0
        for offer_id in offer_id_sendable_sorted:
            if is_lapsed(offer_lookup, offer_id):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr,
                                                                     cluster_rules,
                                                                     offer_lookup,
                                                                     offer_group_lookup,
                                                                     offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "lapsed", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    lapsed_count += 1
                    if lapsed_count >= lapsed_rule:
                        break

    # filter base rules
    offer_id_sendable_sorted = list(filter(
        lambda x: (x not in topk_offer_id) and (not is_lapsed(offer_lookup, x)),
        offer_id_sendable_sorted
    ))
    if topk_offer_attr["bundle"] < 1:
        for offer_id in offer_id_sendable_sorted:
            if is_structure_bundle(offer_lookup, offer_id):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr,
                                                                     cluster_rules,
                                                                     offer_lookup,
                                                                     offer_group_lookup,
                                                                     offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "bundle", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    break
    offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
    if topk_offer_attr["sides"] < 1:
        for offer_id in offer_id_sendable_sorted:
            if is_structure_sides(offer_lookup, offer_id):
                satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr,
                                                                     cluster_rules,
                                                                     offer_lookup,
                                                                     offer_group_lookup,
                                                                     offer_id,
                                                                     use_cluster_rules=True)
                if satisfy_rule:
                    # print(offer_id, "sides", topk_offer_attr)
                    topk_offer_id.append(offer_id)
                    break

    # filter cluster rules
    offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
    for offer_id in offer_id_sendable_sorted:
        satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr,
                                                             cluster_rules,
                                                             offer_lookup,
                                                             offer_group_lookup,
                                                             offer_id,
                                                             use_cluster_rules=True)
        if satisfy_rule:
            # print(offer_id, "cluster", topk_offer_attr)
            topk_offer_id.append(offer_id)
            if len(topk_offer_id) >= topk:
                break

    # fill rest of the array
    if len(topk_offer_id) < topk:
        offer_id_sendable_sorted = list(filter(lambda x: x not in topk_offer_id, offer_id_sendable_sorted))
        for offer_id in offer_id_sendable_sorted:
            satisfy_rule, topk_offer_attr = update_cluster_rules(topk_offer_attr,
                                                                 cluster_rules,
                                                                 offer_lookup,
                                                                 offer_group_lookup,
                                                                 offer_id,
                                                                 use_cluster_rules=False)
            if satisfy_rule:
                # print(offer_id, "cluster", topk_offer_attr)
                topk_offer_id.append(offer_id)
                if len(topk_offer_id) >= topk:
                    break

    return topk_offer_id, transform_offer_attr(topk_offer_attr)


def filter_by_offer_groups(available_offers, selected_offers, offer_group_lookup):
    selected_offers_groups = [offer_group_lookup[offer_id] for offer_id in selected_offers]
    filtered_available_offers = [
        offer_id for offer_id in available_offers
        if offer_group_lookup[offer_id] == 0 or offer_group_lookup[offer_id] not in selected_offers_groups
    ]
    return filtered_available_offers


def get_delivery_offer(delivery_offers, selected_offers, offer_group_lookup):
    filtered_available_offers = filter_by_offer_groups(available_offers=delivery_offers,
                                                       selected_offers=selected_offers,
                                                       offer_group_lookup=offer_group_lookup)
    personalized_delivery_offers = [random.choice(filtered_available_offers)]
    personalized_delivery_offer_attr = {
        "groups": [offer_group_lookup[offer] for offer in personalized_delivery_offers]
    }
    return personalized_delivery_offers, str(personalized_delivery_offer_attr)
